import unittest
import copy
import pathlib

from demes import (
    Epoch,
    Migration,
    Pulse,
    Deme,
    DemeGraph,
    Split,
    Branch,
    Merge,
    Admix,
    load,
)

import demes


class TestEpoch(unittest.TestCase):
    def test_bad_time(self):
        for start_time in (-10000, -1, -1e-9):
            with self.assertRaises(ValueError):
                Epoch(start_time=start_time, end_time=0, initial_size=1)
        for end_time in (-10000, -1, -1e-9, float("inf")):
            with self.assertRaises(ValueError):
                Epoch(start_time=100, end_time=end_time, initial_size=1)

    def test_bad_time_span(self):
        with self.assertRaises(ValueError):
            Epoch(start_time=1, end_time=1, initial_size=1)
        with self.assertRaises(ValueError):
            Epoch(start_time=1, end_time=2, initial_size=1)

    def test_bad_size(self):
        for size in (-10000, -1, -1e-9, 0, float("inf")):
            with self.assertRaises(ValueError):
                Epoch(start_time=1, end_time=0, initial_size=size)
            with self.assertRaises(ValueError):
                Epoch(start_time=1, end_time=0, initial_size=1, final_size=size)

    def test_missing_size(self):
        with self.assertRaises(ValueError):
            Epoch(start_time=1, end_time=0)

    def test_valid_epochs(self):
        Epoch(end_time=0, initial_size=1)
        Epoch(end_time=0, final_size=1)
        Epoch(start_time=float("inf"), end_time=0, initial_size=1)
        Epoch(start_time=float("inf"), end_time=10, initial_size=1)
        Epoch(start_time=100, end_time=99, initial_size=1)
        Epoch(end_time=0, initial_size=1, final_size=1)
        Epoch(end_time=0, initial_size=1, final_size=100)
        Epoch(end_time=0, initial_size=100, final_size=1)
        Epoch(start_time=20, end_time=10, initial_size=1, final_size=100)

    def test_inf_start_time_constant_epoch(self):
        with self.assertRaises(ValueError):
            Epoch(start_time=float("inf"), end_time=0, initial_size=10, final_size=20)

    def test_isclose(self):
        eps = 1e-50
        e1 = Epoch(end_time=0, initial_size=1)
        self.assertTrue(e1.isclose(e1))
        self.assertTrue(e1.isclose(Epoch(end_time=0 + eps, initial_size=1)))
        self.assertTrue(e1.isclose(Epoch(end_time=0, initial_size=1 + eps)))

        self.assertFalse(e1.isclose(Epoch(end_time=1e-9, initial_size=1)))
        self.assertFalse(e1.isclose(Epoch(end_time=0, initial_size=1 + 1e-9)))
        self.assertFalse(e1.isclose(Epoch(start_time=10, end_time=0, initial_size=1)))
        self.assertFalse(e1.isclose(Epoch(end_time=0, initial_size=1, final_size=2)))
        self.assertFalse(
            Epoch(end_time=0, initial_size=1, final_size=2).isclose(
                Epoch(
                    end_time=0,
                    initial_size=1,
                    final_size=2,
                    size_function="exponential",
                )
            )
        )
        self.assertFalse(
            e1.isclose(Epoch(end_time=0, initial_size=1, selfing_rate=0.1))
        )
        self.assertFalse(
            e1.isclose(Epoch(end_time=0, initial_size=1, cloning_rate=0.1))
        )
        self.assertFalse(e1.isclose(None))
        self.assertFalse(e1.isclose(123))
        self.assertFalse(e1.isclose("foo"))

    # APR (7/28): Add tests for selfing rate, cloning rate, and size function.


class TestMigration(unittest.TestCase):
    def test_bad_time(self):
        for time in (-10000, -1, -1e-9):
            with self.assertRaises(ValueError):
                Migration("a", "b", start_time=time, end_time=0, rate=0.1)
        for time in (-10000, -1, -1e-9, float("inf")):
            with self.assertRaises(ValueError):
                Migration("a", "b", start_time=100, end_time=time, rate=0.1)

    def test_bad_rate(self):
        for rate in (-10000, -1, -1e-9, float("inf")):
            with self.assertRaises(ValueError):
                Migration("a", "b", start_time=10, end_time=0, rate=rate)

    def test_bad_demes(self):
        with self.assertRaises(ValueError):
            Migration("a", "a", start_time=10, end_time=0, rate=0.1)

    def test_valid_migration(self):
        Migration("a", "b", start_time=float("inf"), end_time=0, rate=1e-9)
        Migration("a", "b", start_time=1000, end_time=999, rate=0.9)

    def test_isclose(self):
        eps = 1e-50
        m1 = Migration("a", "b", start_time=1, end_time=0, rate=1e-9)
        self.assertTrue(m1.isclose(m1))
        self.assertTrue(
            m1.isclose(Migration("a", "b", start_time=1, end_time=0, rate=1e-9 + eps))
        )
        self.assertTrue(
            m1.isclose(Migration("a", "b", start_time=1 + eps, end_time=0, rate=1e-9))
        )
        self.assertTrue(
            m1.isclose(Migration("a", "b", start_time=1, end_time=0 + eps, rate=1e-9))
        )

        self.assertFalse(
            m1.isclose(Migration("b", "a", start_time=1, end_time=0, rate=1e-9))
        )
        self.assertFalse(
            m1.isclose(Migration("a", "b", start_time=1, end_time=0, rate=2e-9))
        )
        self.assertFalse(
            m1.isclose(Migration("a", "c", start_time=1, end_time=0, rate=1e-9))
        )
        self.assertFalse(
            m1.isclose(Migration("a", "c", start_time=2, end_time=0, rate=1e-9))
        )
        self.assertFalse(
            m1.isclose(Migration("a", "c", start_time=1, end_time=0.1, rate=1e-9))
        )
        self.assertFalse(m1.isclose(None))
        self.assertFalse(m1.isclose(123))
        self.assertFalse(m1.isclose("foo"))


class TestPulse(unittest.TestCase):
    def test_bad_time(self):
        for time in (-10000, -1, -1e-9, float("inf")):
            with self.assertRaises(ValueError):
                Pulse("a", "b", time=time, proportion=0.1)

    def test_bad_proportion(self):
        for proportion in (-10000, -1, -1e-9, 1.2, 100, float("inf")):
            with self.assertRaises(ValueError):
                Pulse("a", "b", time=1, proportion=proportion)

    def test_bad_demes(self):
        with self.assertRaises(ValueError):
            Pulse("a", "a", time=1, proportion=0.1)

    def test_valid_pulse(self):
        Pulse("a", "b", time=1, proportion=1e-9)
        Pulse("a", "b", time=100, proportion=0.9)

    def test_isclose(self):
        eps = 1e-50
        p1 = Pulse("a", "b", time=1, proportion=1e-9)
        self.assertTrue(p1.isclose(p1))
        self.assertTrue(p1.isclose(Pulse("a", "b", time=1, proportion=1e-9)))
        self.assertTrue(p1.isclose(Pulse("a", "b", time=1 + eps, proportion=1e-9)))
        self.assertTrue(p1.isclose(Pulse("a", "b", time=1, proportion=1e-9 + eps)))

        self.assertFalse(p1.isclose(Pulse("a", "c", time=1, proportion=1e-9)))
        self.assertFalse(p1.isclose(Pulse("b", "a", time=1, proportion=1e-9)))
        self.assertFalse(p1.isclose(Pulse("a", "b", time=1, proportion=2e-9)))
        self.assertFalse(p1.isclose(Pulse("a", "b", time=1 + 1e-9, proportion=1e-9)))


class TestSplit(unittest.TestCase):
    def test_bad_time(self):
        for time in [-1e-12, -1, float("inf")]:
            with self.assertRaises(ValueError):
                Split("a", ["b", "c"], time)

    def test_children(self):
        with self.assertRaises(ValueError):
            Split("a", "b", 1)
        with self.assertRaises(ValueError):
            Split("a", ["a", "b"], 1)

    def test_valid_split(self):
        Split("a", ["b", "c"], 10)
        Split("a", ["b", "c", "d"], 10)
        Split("a", ["b", "c"], 0)

    def test_isclose(self):
        eps = 1e-50
        s1 = Split("a", ["b", "c"], 1)
        self.assertTrue(s1.isclose(s1))
        self.assertTrue(s1.isclose(Split("a", ["b", "c"], 1)))
        self.assertTrue(s1.isclose(Split("a", ["b", "c"], 1 + eps)))
        # Order of children doesn't matter.
        self.assertTrue(s1.isclose(Split("a", ["c", "b"], 1)))

        self.assertFalse(s1.isclose(Split("a", ["x", "c"], 1)))
        self.assertFalse(s1.isclose(Split("x", ["b", "c"], 1)))
        self.assertFalse(s1.isclose(Split("a", ["b", "c", "x"], 1)))
        self.assertFalse(s1.isclose(Split("a", ["b", "c"], 1 + 1e-9)))


class TestBranch(unittest.TestCase):
    def test_bad_time(self):
        for time in [-1e-12, -1, float("inf")]:
            with self.assertRaises(ValueError):
                Branch("a", "b", time)

    def test_branch_demes(self):
        with self.assertRaises(ValueError):
            Branch("a", "a", 1)

    def test_valid_branch(self):
        Branch("a", "b", 10)
        Branch("a", "b", 0)

    def test_isclose(self):
        eps = 1e-50
        b1 = Branch("a", "b", 1)
        self.assertTrue(b1.isclose(b1))
        self.assertTrue(b1.isclose(Branch("a", "b", 1)))
        self.assertTrue(b1.isclose(Branch("a", "b", 1 + eps)))

        self.assertFalse(b1.isclose(Branch("x", "b", 1)))
        self.assertFalse(b1.isclose(Branch("a", "x", 1)))
        self.assertFalse(b1.isclose(Branch("b", "a", 1)))
        self.assertFalse(b1.isclose(Branch("a", "b", 1 + 1e-9)))


class TestMerge(unittest.TestCase):
    def test_bad_time(self):
        for time in [-1e-12, -1, float("inf")]:
            with self.assertRaises(ValueError):
                Merge(["a", "b"], [0.5, 0.5], "c", time)

    def test_bad_parents_proportions(self):
        with self.assertRaises(ValueError):
            Merge("a", [1], "b", 1)
        with self.assertRaises(ValueError):
            Merge(["a"], 1.0, "b", 10)
        with self.assertRaises(ValueError):
            Merge(["a"], [1], "b", 1)
        with self.assertRaises(ValueError):
            Merge(["a", "b"], [0.5, 0.5], "a", 1)
        with self.assertRaises(ValueError):
            Merge(["a", "a"], [0.5, 0.5], "b", 1)

    def test_invalid_proportions(self):
        with self.assertRaises(ValueError):
            Merge(["a", "b"], [0.1, 1], "c", 1)
        with self.assertRaises(ValueError):
            Merge(["a", "b"], [0.5], "c", 1)
        with self.assertRaises(ValueError):
            Merge(["a", "b"], [1.0], "c", 1)
        with self.assertRaises(ValueError):
            Merge(["a", "b", "c"], [0.5, 0.5, 0.5], "d", 1)

    def test_valid_merge(self):
        Merge(["a", "b"], [0.5, 0.5], "c", 10)
        Merge(["a", "b"], [0.5, 0.5], "c", 0)
        Merge(["a", "b", "c"], [0.5, 0.25, 0.25], "d", 10)
        Merge(["a", "b", "c"], [0.5, 0.5, 0.0], "d", 10)
        Merge(["a", "b"], [1, 0], "c", 10)

    def test_isclose(self):
        eps = 1e-50
        m1 = Merge(["a", "b"], [0.1, 0.9], "c", 1)
        self.assertTrue(m1.isclose(m1))
        self.assertTrue(m1.isclose(Merge(["a", "b"], [0.1, 0.9], "c", 1)))
        self.assertTrue(m1.isclose(Merge(["a", "b"], [0.1, 0.9], "c", 1 + eps)))
        self.assertTrue(m1.isclose(Merge(["a", "b"], [0.1 + eps, 0.9], "c", 1)))
        self.assertTrue(m1.isclose(Merge(["a", "b"], [0.1, 0.9 + eps], "c", 1)))
        # Order of parents/proportions doesn't matter.
        self.assertTrue(m1.isclose(Merge(["b", "a"], [0.9, 0.1], "c", 1)))

        self.assertFalse(m1.isclose(Merge(["a", "x"], [0.1, 0.9], "c", 1)))
        self.assertFalse(m1.isclose(Merge(["x", "b"], [0.1, 0.9], "c", 1)))
        self.assertFalse(
            m1.isclose(Merge(["a", "b"], [0.1 + 1e-9, 0.9 - 1e-9], "c", 1))
        )
        self.assertFalse(m1.isclose(Merge(["a", "b"], [0.1, 0.9], "x", 1)))
        self.assertFalse(m1.isclose(Merge(["a", "b"], [0.1, 0.9], "c", 1 + 1e-9)))
        self.assertFalse(m1.isclose(Merge(["a", "b", "x"], [0.1, 0.9, 0], "c", 1)))


class TestAdmix(unittest.TestCase):
    def test_bad_time(self):
        for time in [-1e-12, -1, float("inf")]:
            with self.assertRaises(ValueError):
                Admix(["a", "b"], [0.5, 0.5], "c", time)

    def test_bad_parents_proportions(self):
        with self.assertRaises(ValueError):
            Admix("a", [1], "b", 1)
        with self.assertRaises(ValueError):
            Admix(["a"], 1.0, "b", 10)
        with self.assertRaises(ValueError):
            Admix(["a"], [1], "b", 1)
        with self.assertRaises(ValueError):
            Admix(["a", "b"], [0.5, 0.5], "a", 1)
        with self.assertRaises(ValueError):
            Admix(["a", "a"], [0.5, 0.5], "b", 1)

    def test_invalid_proportions(self):
        with self.assertRaises(ValueError):
            Admix(["a", "b"], [0.1, 1], "c", 1)
        with self.assertRaises(ValueError):
            Admix(["a", "b"], [0.5], "c", 1)
        with self.assertRaises(ValueError):
            Admix(["a", "b"], [1.0], "c", 1)
        with self.assertRaises(ValueError):
            Admix(["a", "b", "c"], [0.5, 0.5, 0.5], "d", 1)

    def test_valid_admixture(self):
        Admix(["a", "b"], [0.5, 0.5], "c", 10)
        Admix(["a", "b"], [0.5, 0.5], "c", 0)
        Admix(["a", "b", "c"], [0.5, 0.25, 0.25], "d", 10)
        Admix(["a", "b", "c"], [0.5, 0.5, 0.0], "d", 10)
        Admix(["a", "b"], [1, 0], "c", 10)

    def test_isclose(self):
        eps = 1e-50
        a1 = Admix(["a", "b"], [0.1, 0.9], "c", 1)
        self.assertTrue(a1.isclose(a1))
        self.assertTrue(a1.isclose(Admix(["a", "b"], [0.1, 0.9], "c", 1)))
        self.assertTrue(a1.isclose(Admix(["a", "b"], [0.1 + eps, 0.9], "c", 1 + eps)))
        self.assertTrue(a1.isclose(Admix(["a", "b"], [0.1 + eps, 0.9], "c", 1)))
        self.assertTrue(a1.isclose(Admix(["a", "b"], [0.1, 0.9 + eps], "c", 1 + eps)))
        # Order of parents/proportions doesn't matter.
        self.assertTrue(a1.isclose(Admix(["b", "a"], [0.9, 0.1], "c", 1)))

        self.assertFalse(a1.isclose(Admix(["a", "x"], [0.1, 0.9], "c", 1)))
        self.assertFalse(a1.isclose(Admix(["x", "b"], [0.1, 0.9], "c", 1)))
        self.assertFalse(
            a1.isclose(Admix(["a", "b"], [0.1 + 1e-9, 0.9 - 1e-9], "c", 1))
        )
        self.assertFalse(a1.isclose(Admix(["a", "b"], [0.1, 0.9], "x", 1)))
        self.assertFalse(a1.isclose(Admix(["a", "b"], [0.1, 0.9], "c", 1 + 1e-9)))
        self.assertFalse(a1.isclose(Admix(["a", "b", "x"], [0.1, 0.9, 0], "c", 1)))


class TestDeme(unittest.TestCase):
    def test_properties(self):
        deme = Deme(
            "a",
            "b",
            ["c"],
            [1],
            [Epoch(start_time=float("inf"), end_time=0, initial_size=1)],
        )
        self.assertEqual(deme.start_time, float("inf"))
        self.assertEqual(deme.end_time, 0)
        self.assertEqual(deme.ancestors[0], "c")
        self.assertEqual(deme.proportions[0], 1)

        deme = Deme(
            "a", "b", ["c"], [1], [Epoch(start_time=100, end_time=50, initial_size=1)]
        )
        self.assertEqual(deme.start_time, 100)
        self.assertEqual(deme.end_time, 50)
        deme.add_epoch(Epoch(start_time=50, end_time=20, initial_size=100))
        self.assertEqual(deme.start_time, 100)
        self.assertEqual(deme.end_time, 20)
        deme.add_epoch(Epoch(start_time=20, end_time=1, initial_size=200))
        self.assertEqual(deme.start_time, 100)
        self.assertEqual(deme.end_time, 1)

    def test_no_epochs(self):
        with self.assertRaises(ValueError):
            Deme("a", "b", ["c"], [1], [])

    def test_bad_ancestors(self):
        with self.assertRaises(TypeError):
            Deme(
                "a",
                "b",
                "c",
                [1],
                [Epoch(start_time=10, end_time=0, initial_size=1)],
            )
        with self.assertRaises(TypeError):
            Deme(
                "a",
                "b",
                {"c", "d"},
                [0.2, 0.8],
                [Epoch(start_time=10, end_time=0, initial_size=1)],
            )
        with self.assertRaises(ValueError):
            Deme(
                "a",
                "b",
                ["c", "d"],
                None,
                [Epoch(start_time=10, end_time=0, initial_size=1)],
            )
        with self.assertRaises(ValueError):
            Deme(
                "a",
                "b",
                ["c", "d"],
                [0.5, 0.2, 0.3],
                [Epoch(start_time=10, end_time=0, initial_size=1)],
            )
        with self.assertRaises(ValueError):
            Deme(
                "a",
                "b",
                ["a", "c"],
                [0.5, 0.5],
                [Epoch(start_time=10, end_time=0, initial_size=1)],
            )

    def test_two_epochs(self):
        with self.assertRaises(ValueError):
            Deme(
                "a",
                "b",
                ["c"],
                [1],
                [
                    Epoch(start_time=11, end_time=10, initial_size=100),
                    Epoch(start_time=10, end_time=1, initial_size=1),
                ],
            )

    def test_add_epoch(self):
        deme = Deme(
            "a", "b", ["c"], [1], [Epoch(start_time=100, end_time=50, initial_size=1)]
        )
        with self.assertRaises(ValueError):
            deme.add_epoch(Epoch(start_time=60, end_time=0, initial_size=10))
        with self.assertRaises(ValueError):
            deme.add_epoch(Epoch(start_time=40, end_time=0, initial_size=10))
        with self.assertRaises(ValueError):
            deme.add_epoch(Epoch(start_time=50, end_time=50, initial_size=10))
        with self.assertRaises(ValueError):
            deme.add_epoch(
                Epoch(
                    start_time=50,
                    end_time=0,
                    initial_size=10,
                    final_size=20,
                    size_function="constant",
                )
            )

    def test_epochs_out_of_order(self):
        deme = Deme(
            "a", "b", ["c"], [1], [Epoch(start_time=10, end_time=5, initial_size=1)]
        )
        for time in (5, -1, float("inf")):
            with self.assertRaises(ValueError):
                deme.add_epoch(Epoch(start_time=5, end_time=time, initial_size=100))
        deme.add_epoch(Epoch(start_time=5, end_time=0, initial_size=100))

    def test_epochs_are_a_partition(self):
        for start_time, end_time in [(float("inf"), 100), (200, 100)]:
            deme = Deme(
                "a",
                "b",
                ["c"],
                [1],
                [Epoch(start_time=start_time, end_time=end_time, initial_size=1)],
            )
            with self.assertRaises(ValueError):
                deme.add_epoch(Epoch(end_time=100, initial_size=100))
            for t in (50, 20, 10):
                deme.add_epoch(Epoch(end_time=t, initial_size=t))
            prev_end_time = end_time
            for epoch in deme.epochs[1:]:
                self.assertEqual(epoch.start_time, prev_end_time)
                prev_end_time = epoch.end_time

    def test_time_span(self):
        for start_time, end_time in zip((float("inf"), 100, 20), (0, 20, 0)):
            deme = Deme(
                "a",
                "b",
                ["c"],
                [1],
                [Epoch(start_time=start_time, end_time=end_time, initial_size=1)],
            )
            self.assertEqual(deme.time_span, start_time - end_time)

    def test_isclose(self):
        d1 = Deme(
            id="a",
            description="foo deme",
            ancestors=None,
            proportions=None,
            epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
        )
        self.assertTrue(d1.isclose(d1))
        self.assertTrue(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
                )
            )
        )
        # Description field doesn't matter.
        self.assertTrue(
            d1.isclose(
                Deme(
                    id="a",
                    description="bar deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
                )
            )
        )

        # selfing_rate is a property of the deme's epoch, so we shouldn't
        # care if this is set for the deme or for the epochs directly.
        d2 = Deme(
            id="a",
            description="foo deme",
            ancestors=None,
            proportions=None,
            epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
            selfing_rate=0.1,
        )
        self.assertTrue(
            d2.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[
                        Epoch(
                            start_time=10, end_time=5, initial_size=1, selfing_rate=0.1
                        )
                    ],
                )
            )
        )

        # cloning_rate is a property of the deme's epoch, so we shouldn't
        # care if this is set for the deme or for the epochs directly.
        d2 = Deme(
            id="a",
            description="foo deme",
            ancestors=None,
            proportions=None,
            epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
            cloning_rate=0.1,
        )
        self.assertTrue(
            d2.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[
                        Epoch(
                            start_time=10, end_time=5, initial_size=1, cloning_rate=0.1
                        )
                    ],
                )
            )
        )

        #
        # Check inequalities.
        #

        self.assertFalse(
            d1.isclose(
                Deme(
                    id="b",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
                )
            )
        )
        self.assertFalse(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=["x"],
                    proportions=[1],
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
                )
            )
        )
        self.assertFalse(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=9, end_time=5, initial_size=1)],
                )
            )
        )
        self.assertFalse(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=9, initial_size=1)],
                )
            )
        )

        self.assertFalse(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=9)],
                )
            )
        )
        self.assertFalse(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
                    selfing_rate=0.1,
                )
            )
        )
        self.assertFalse(
            d1.isclose(
                Deme(
                    id="a",
                    description="foo deme",
                    ancestors=None,
                    proportions=None,
                    epochs=[Epoch(start_time=10, end_time=5, initial_size=1)],
                    cloning_rate=0.1,
                )
            )
        )
        d2 = copy.deepcopy(d1)
        d2.add_epoch(Epoch(end_time=0, initial_size=1))
        self.assertFalse(d1 == d2)

    # APR (7/28): Add tests for selfing rate, cloning rate, and size function.
    # Add tests for testing ancestors and proportions.
    # Also add tests for any implied values.


class TestDemeGraph(unittest.TestCase):
    def test_bad_generation_time(self):
        for generation_time in (-100, -1e-9, 0, float("inf")):
            with self.assertRaises(ValueError):
                DemeGraph(
                    description="test",
                    time_units="years",
                    generation_time=generation_time,
                )

    def test_bad_default_Ne(self):
        for N in (-100, -1e-9, 0, float("inf")):
            with self.assertRaises(ValueError):
                DemeGraph(
                    description="test",
                    time_units="years",
                    generation_time=1,
                    default_Ne=N,
                )

    def check_in_generations(self, dg1):
        assert dg1.generation_time is not None
        assert dg1.generation_time > 1
        dg1_copy = copy.deepcopy(dg1)
        dg2 = dg1.in_generations()
        # in_generations() shouldn't modify the original
        self.assertEqual(dg1.asdict(), dg1_copy.asdict())
        # but clearly dg2 should now differ
        self.assertNotEqual(dg1.asdict(), dg2.asdict())

        # Alternate implementation, which recurses the object hierarchy.
        def in_generations2(dg):
            generation_time = dg.generation_time
            dg = copy.deepcopy(dg)
            dg.time_units = "generations"
            if generation_time is None:
                return dg
            dg.generation_time = None

            def divide_time_attrs(obj):
                if not hasattr(obj, "__dict__"):
                    return
                for name, value in obj.__dict__.items():
                    if name in ("time", "start_time", "end_time"):
                        if value is not None:
                            setattr(obj, name, value / generation_time)
                    elif isinstance(value, (list, tuple)):
                        for a in value:
                            divide_time_attrs(a)
                    else:
                        divide_time_attrs(value)

            divide_time_attrs(dg)
            return dg

        self.assertEqual(in_generations2(dg1).asdict(), dg2.asdict())
        # in_generations2() shouldn't modify the original
        self.assertEqual(dg1.asdict(), dg1_copy.asdict())

        # in_generations() should be idempotent
        dg3 = dg2.in_generations()
        self.assertEqual(dg2.asdict(), dg3.asdict())
        dg3 = in_generations2(dg2)
        self.assertEqual(dg2.asdict(), dg3.asdict())

    def test_in_generations(self):
        examples = list(pathlib.Path("examples").glob("*.yml"))
        examples.extend(list(pathlib.Path("../examples").glob("*.yml")))
        if len(examples) == 0:
            raise RuntimeError(
                "Can't find any examples. The tests should be run from the "
                "'test' directory or the toplevel directory."
            )
        i = 0
        for yml in examples:
            dg = load(yml)
            if dg.generation_time not in (None, 1):
                self.check_in_generations(dg)
                i += 1
        self.assertGreater(i, 0)

    def test_bad_migration_time(self):
        dg = demes.DemeGraph(description="test bad migration", time_units="generations")
        dg.deme("deme1", end_time=0, initial_size=1000)
        dg.deme("deme2", end_time=100, initial_size=1000)
        with self.assertRaises(ValueError):
            dg.migration(
                source="deme1", dest="deme2", rate=0.01, start_time=1000, end_time=0
            )

    def test_bad_pulse_time(self):
        dg = demes.DemeGraph(
            description="test bad pulse time", time_units="generations"
        )
        dg.deme("deme1", end_time=0, initial_size=1000)
        dg.deme("deme2", end_time=100, initial_size=1000)
        with self.assertRaises(ValueError):
            dg.pulse(source="deme1", dest="deme2", proportion=0.1, time=10)

    def test_bad_deme(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        with self.assertRaises(ValueError):
            dg.deme("a")
        with self.assertRaises(TypeError):
            dg.deme("a", initial_size=100, ancestors="b")
        with self.assertRaises(ValueError):
            dg.deme("a", initial_size=100, ancestors=["b", "c"], proportions=[0.5, 0.5])
        with self.assertRaises(ValueError):
            dg.deme(
                "a",
                initial_size=100,
                ancestors=["b"],
                end_time=0,
                epochs=[
                    Epoch(initial_size=1, start_time=20, end_time=10),
                    Epoch(start_time=10, end_time=5, initial_size=2),
                ],
            )
        with self.assertRaises(ValueError):
            dg.deme(
                "a",
                initial_size=100,
                ancestors=["b"],
                start_time=15,
                end_time=0,
                epochs=[
                    Epoch(initial_size=1, start_time=20, end_time=10),
                    Epoch(start_time=10, end_time=0, initial_size=2),
                ],
            )

    def test_proportions_default(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        with self.assertRaises(ValueError):
            dg.deme("a", initial_size=100, ancestors=["b", "c"])
        dg.deme("a", initial_size=100, ancestors=["b"])
        self.assertEqual(len(dg["a"].proportions), 1)
        self.assertEqual(dg["a"].proportions[0], 1.0)

    def test_bad_migration(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        with self.assertRaises(ValueError):
            dg.symmetric_migration()
        with self.assertRaises(ValueError):
            dg.symmetric_migration(demes=["a"], rate=0.1)
        with self.assertRaises(ValueError):
            dg.migration("a", "b", rate=0.1)
        dg.deme("a", initial_size=100)
        with self.assertRaises(ValueError):
            dg.migration("a", "b", rate=0.1)
        with self.assertRaises(ValueError):
            dg.migration("b", "a", rate=0.1)

    def test_bad_pulse(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", initial_size=100)
        with self.assertRaises(ValueError):
            dg.pulse("a", "b", 0.1, 10)
        with self.assertRaises(ValueError):
            dg.pulse("b", "a", 0.1, 10)

    def test_bad_split(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        with self.assertRaises(ValueError):
            dg.split("a", ["a", "b", "c"], 10)
        dg.deme("a", initial_size=100, end_time=50)
        dg.deme("b", initial_size=100, start_time=50, end_time=0)
        dg.deme("c", initial_size=100, start_time=20, end_time=0)
        with self.assertRaises(ValueError):
            dg.split("a", ["b", "c"], 50)

    def test_bad_branch(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", start_time=10, end_time=5, initial_size=10)
        dg.deme("b", start_time=20, end_time=0, initial_size=10)
        dg.deme("c", start_time=2, end_time=0, initial_size=10)
        with self.assertRaises(ValueError):
            dg.branch("a", "b", 7)
        with self.assertRaises(ValueError):
            dg.branch("a", "c", 7)

    def test_bad_merge(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", end_time=5, initial_size=10)
        dg.deme("b", end_time=5, initial_size=10)
        dg.deme("c", start_time=5, end_time=0, initial_size=10)
        dg.deme("d", start_time=2, end_time=0, initial_size=1)
        with self.assertRaises(ValueError):
            dg.merge(["a", "b"], [0.5, 0.5], "c", 10)
        with self.assertRaises(ValueError):
            dg.merge(["a", "b"], [0.5, 0.5], "d", 2)

    def test_merge_cuts_epochs(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", initial_size=10, end_time=0)
        dg.deme(
            "b",
            epochs=[
                Epoch(end_time=10, initial_size=5),
                Epoch(end_time=5, initial_size=10),
                Epoch(end_time=0, initial_size=2),
            ],
        )
        dg.deme("c", start_time=5, initial_size=10)
        dg.merge(["a", "b"], [0.5, 0.5], "c", 5)
        self.assertEqual(dg["a"].end_time, 5)
        self.assertEqual(dg["b"].end_time, 5)
        self.assertEqual(len(dg["b"].epochs), 2)

    def test_bad_admixture(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", start_time=2, initial_size=100)
        with self.assertRaises(ValueError):
            dg.admix(["b", "c"], [0.5, 0.5], "a", 5)
        dg.deme("b", end_time=5, initial_size=10)
        dg.deme("c", start_time=5, end_time=0, initial_size=10)
        with self.assertRaises(ValueError):
            dg.admix(["b", "c"], [0.5, 0.5], "a", 2)

    def test_isclose(self):
        g1 = DemeGraph(
            description="test",
            time_units="generations",
        )
        g2 = copy.deepcopy(g1)
        g1.deme("d1", initial_size=1000)
        self.assertTrue(g1.isclose(g1))
        self.assertTrue(g1.isclose(demes.loads(demes.dumps(g1))))

        # Don't care about description for equality.
        g3 = DemeGraph(
            description="some other description",
            time_units="generations",
        )
        g3.deme("d1", initial_size=1000)
        self.assertTrue(g1.isclose(g3))

        # Don't care about doi for equality.
        g3 = DemeGraph(
            description="test",
            time_units="generations",
            doi="https://example.com/foo.bar",
        )
        g3.deme("d1", initial_size=1000)
        self.assertTrue(g1.isclose(g3))

        # The choice of using default_Ne, or alternately setting initial_size
        # for each deme, is a model implementation detail. So the resulting
        # deme graphs should compare equal.
        g3 = DemeGraph(
            description="test",
            time_units="generations",
            default_Ne=1000,
        )
        g3.deme("d1")
        self.assertTrue(g1.isclose(g3))

        # Selfing rate is a property of a deme's epoch, not a deme graph.
        g3 = DemeGraph(
            description="test",
            time_units="generations",
            selfing_rate=0.1,
        )
        g3.deme("d1", initial_size=1000)
        g4 = DemeGraph(
            description="test",
            time_units="generations",
        )
        g4.deme("d1", initial_size=1000, selfing_rate=0.1)
        self.assertTrue(g3.isclose(g4))

        # Cloning rate is a property of a deme's epoch, not a deme graph.
        g3 = DemeGraph(
            description="test",
            time_units="generations",
            cloning_rate=0.1,
        )
        g3.deme("d1", initial_size=1000)
        g4 = DemeGraph(
            description="test",
            time_units="generations",
        )
        g4.deme("d1", initial_size=1000, cloning_rate=0.1)
        self.assertTrue(g3.isclose(g4))

        # The order in which demes are added shouldn't matter.
        g3 = copy.deepcopy(g2)
        g4 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        g4.deme("d2", initial_size=1000)
        g4.deme("d1", initial_size=1000)
        self.assertTrue(g3.isclose(g4))

        # The order in which migrations are added shouldn't matter.
        g3.migration("d1", "d2", rate=1e-4, start_time=50, end_time=40)
        g3.migration("d2", "d1", rate=1e-5)
        g4.migration("d2", "d1", rate=1e-5)
        g4.migration("d1", "d2", rate=1e-4, start_time=50, end_time=40)
        self.assertTrue(g3.isclose(g4))

        # The order in which pulses are added shouldn't matter.
        g3.pulse("d1", "d2", proportion=0.01, time=100)
        g3.pulse("d1", "d2", proportion=0.01, time=50)
        g4.pulse("d1", "d2", proportion=0.01, time=50)
        g4.pulse("d1", "d2", proportion=0.01, time=100)
        self.assertTrue(g3.isclose(g4))

        #
        # Check inequalities
        #

        self.assertFalse(g1 == g2)
        g3 = copy.deepcopy(g2)
        g3.deme("dX", initial_size=1000)
        self.assertFalse(g1.isclose(g3))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1001)
        self.assertFalse(g1.isclose(g3))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        self.assertFalse(g1.isclose(g3))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", ancestors=["x"], initial_size=1000)
        self.assertFalse(g1.isclose(g3))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        g4 = copy.deepcopy(g2)
        g4.deme("d1", initial_size=1000)
        g4.deme("d2", initial_size=1000)
        g4.migration("d2", "d1", rate=1e-5)
        self.assertFalse(g3.isclose(g4))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        g3.migration("d1", "d2", rate=1e-5)
        g4 = copy.deepcopy(g2)
        g4.deme("d1", initial_size=1000)
        g4.deme("d2", initial_size=1000)
        g4.migration("d2", "d1", rate=1e-5)
        self.assertFalse(g3.isclose(g4))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        g3.migration("d2", "d1", rate=1e-5)
        g4 = copy.deepcopy(g2)
        g4.deme("d1", initial_size=1000)
        g4.deme("d2", initial_size=1000)
        g4.symmetric_migration(["d2", "d1"], rate=1e-5)
        self.assertFalse(g3.isclose(g4))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        g4 = copy.deepcopy(g2)
        g4.deme("d1", initial_size=1000)
        g4.deme("d2", initial_size=1000)
        g4.pulse("d1", "d2", proportion=0.01, time=100)
        self.assertFalse(g3.isclose(g4))

        g3 = copy.deepcopy(g2)
        g3.deme("d1", initial_size=1000)
        g3.deme("d2", initial_size=1000)
        g3.pulse("d2", "d1", proportion=0.01, time=100)
        g4 = copy.deepcopy(g2)
        g4.deme("d1", initial_size=1000)
        g4.deme("d2", initial_size=1000)
        g3.pulse("d1", "d2", proportion=0.01, time=100)
        self.assertFalse(g3.isclose(g4))


class TestDemeGraphToDict(unittest.TestCase):
    def test_fill_selfing_rate(self):
        dg = demes.DemeGraph(
            description="a", time_units="generations", selfing_rate=0.5
        )
        dg.deme("a", initial_size=10)
        d = dg.asdict_compact()
        self.assertTrue(d["selfing_rate"] == dg.selfing_rate)

    def test_fill_cloning_rate(self):
        dg = demes.DemeGraph(
            description="a", time_units="generations", cloning_rate=0.5
        )
        dg.deme("a", initial_size=10)
        d = dg.asdict_compact()
        self.assertTrue(d["cloning_rate"] == dg.cloning_rate)

    def test_finite_start_time(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", initial_size=100, start_time=100)
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["start_time"] == dg["a"].start_time)

    def test_deme_selfing_rate(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", initial_size=100, selfing_rate=0.1)
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["selfing_rate"] == dg["a"].selfing_rate)
        dg = demes.DemeGraph(
            description="a", time_units="generations", selfing_rate=0.01
        )
        dg.deme("a", initial_size=100, selfing_rate=0.1)
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["selfing_rate"] == dg["a"].selfing_rate)

    def test_deme_cloning_rate(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", initial_size=100, cloning_rate=0.1)
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["cloning_rate"] == dg["a"].cloning_rate)

    def test_fill_nonstandard_size_function(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme(
            "a",
            epochs=[
                Epoch(initial_size=1, end_time=10),
                Epoch(final_size=10, size_function="linear", end_time=0),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["epochs"][-1]["size_function"] == "linear")

    def test_fill_epoch_selfing_rates(self):
        dg = demes.DemeGraph(
            description="a", time_units="generations", selfing_rate=0.2
        )
        dg.deme(
            "a",
            epochs=[
                Epoch(initial_size=10, end_time=10, selfing_rate=0.2),
                Epoch(final_size=20, end_time=0, selfing_rate=0.1),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["selfing_rate"] == 0.2)
        self.assertTrue("selfing_rate" not in d["demes"]["a"]["epochs"][0])
        self.assertTrue(d["demes"]["a"]["epochs"][1]["selfing_rate"] == 0.1)

        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme(
            "a",
            selfing_rate=0.2,
            epochs=[
                Epoch(initial_size=10, end_time=10, selfing_rate=0.2),
                Epoch(final_size=20, end_time=0, selfing_rate=0.1),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["selfing_rate"] == 0.2)
        self.assertTrue("selfing_rate" not in d["demes"]["a"]["epochs"][0])
        self.assertTrue(d["demes"]["a"]["epochs"][1]["selfing_rate"] == 0.1)

        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme(
            "a",
            epochs=[
                Epoch(initial_size=10, end_time=10),
                Epoch(final_size=20, end_time=0, selfing_rate=0.1),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["epochs"][1]["selfing_rate"] == 0.1)

    def test_fill_epoch_cloning_rates(self):
        dg = demes.DemeGraph(
            description="a", time_units="generations", cloning_rate=0.2
        )
        dg.deme(
            "a",
            epochs=[
                Epoch(initial_size=10, end_time=10),
                Epoch(final_size=20, end_time=0, cloning_rate=0.1),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["cloning_rate"] == 0.2)
        self.assertTrue(d["demes"]["a"]["epochs"][1]["cloning_rate"] == 0.1)

        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme(
            "a",
            cloning_rate=0.2,
            epochs=[
                Epoch(initial_size=10, end_time=10),
                Epoch(final_size=20, end_time=0, cloning_rate=0.1),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["cloning_rate"] == 0.2)
        self.assertTrue(d["demes"]["a"]["epochs"][1]["cloning_rate"] == 0.1)

        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme(
            "a",
            epochs=[
                Epoch(initial_size=10, end_time=10),
                Epoch(final_size=20, end_time=0, cloning_rate=0.1),
            ],
        )
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["epochs"][1]["cloning_rate"] == 0.1)

    def test_fill_description(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", description="described", initial_size=100)
        d = dg.asdict_compact()
        self.assertTrue(d["demes"]["a"]["description"] == dg["a"].description)

    def test_fill_migration_bounds(self):
        dg = demes.DemeGraph(description="a", time_units="generations")
        dg.deme("a", initial_size=100)
        dg.deme("b", initial_size=100)
        dg.migration("a", "b", 0.01, start_time=20, end_time=10)
        d = dg.asdict_compact()
        self.assertTrue(d["migrations"]["asymmetric"][0]["start_time"] == 20)
        self.assertTrue(d["migrations"]["asymmetric"][0]["end_time"] == 10)



"""
Functions to load and dump yaml formatted descriptions of a demography.
The strictyaml schema defined here follows the DemeGraph construction API.

AR: whole-sale changes to comply with the forwards-in-time construction now
in `demes.py`.
"""
# TODO: add symmetric_migration and subgraph schemas.

from strictyaml import (
    Optional,
    Map,
    MapPattern,
    Float,
    Int,
    Seq,
    Str,
    dirty_load,
    as_document,
    CommaSeparated,
)

import demes

Number = Int() | Float()

_epoch_schema = Map(
    {
        Optional("start_time"): Number,
        "end_time": Number,
        Optional("initial_size"): Number,
        Optional("final_size"): Number,
        Optional("size_function"): Str(),
        Optional("selfing_rate"): Number,
        Optional("cloning_rate"): Number,
    }
)

_asymmetric_migration_schema = Map(
    {
        Optional("start_time"): Number,
        Optional("end_time"): Number,
        "source": Str(),
        "dest": Str(),
        "rate": Float(),
    }
)

_symmetric_migration_schema = Map(
    {
        Optional("start_time"): Number,
        Optional("end_time"): Number,
        "demes": CommaSeparated(Str()),
        "rate": Float(),
    }
)

_pulse_schema = Map(
    {"time": Number, "source": Str(), "dest": Str(), "proportion": Float()}
)

_deme_schema = Map(
    {
        Optional("description"): Str(),
        Optional("ancestors"): CommaSeparated(Str()),
        Optional("proportions"): CommaSeparated(Float()),
        Optional("start_time"): Number,
        Optional("end_time"): Number,
        Optional("initial_size"): Number,
        Optional("final_size"): Number,
        Optional("epochs"): Seq(_epoch_schema),
        Optional("selfing_rate"): Number,
        Optional("cloning_rate"): Number,
    }
)

_deme_graph_schema = Map(
    {
        "description": Str(),
        "time_units": Str(),
        Optional("generation_time"): Number,
        Optional("doi"): Str(),
        Optional("default_Ne"): Number,
        "demes": MapPattern(Str(), _deme_schema),
        Optional("migrations"): Map(
            {
                Optional("symmetric"): Seq(_symmetric_migration_schema),
                Optional("asymmetric"): Seq(_asymmetric_migration_schema),
            }
        ),
        Optional("pulses"): Seq(_pulse_schema),
        Optional("selfing_rate"): Number,
        Optional("cloning_rate"): Number,
    }
)


def loads(string):
    """
    Load a deme graph from a yaml-formatted string.

    :param str string: the ``yaml`` string to be loaded.
    :return: A deme graph.
    :rtype: .DemeGraph

    .. todo:: Describe demes' yaml format. The semantics in the yaml
        follow that of the :class:`.DemeGraph` demography construction
        methods.
    """
    yaml = dirty_load(string, schema=_deme_graph_schema, allow_flow_style=True)
    d = yaml.data  # data dict
    g = demes.DemeGraph(
        description=d.get("description"),
        time_units=d.get("time_units"),
        generation_time=d.get("generation_time"),
        doi=d.get("doi"),
        default_Ne=d.get("default_Ne"),
        selfing_rate=d.get("selfing_rate"),
        cloning_rate=d.get("cloning_rate"),
    )
    for deme_id, deme_dict in d.get("demes", dict()).items():
        if "epochs" in deme_dict:
            deme_dict["epochs"] = [
                demes.Epoch(**epoch_dict) for epoch_dict in deme_dict["epochs"]
            ]
        g.deme(deme_id, **deme_dict)
    for migration_type, migration_dict in d.get("migrations", dict()).items():
        if migration_type == "symmetric":
            for m in migration_dict:
                g.symmetric_migration(**m)
        if migration_type == "asymmetric":
            for m in migration_dict:
                g.migration(**m)
    for pulse_dict in d.get("pulses", []):
        g.pulse(**pulse_dict)
    # add population relationship events to the deme graph
    g.get_demographic_events()
    return g


def load(filename):
    """
    Load a deme graph from a yaml-formatted file.

    :param str filename: the name of the ``yaml`` file to load.
    :return: A deme graph.
    :rtype: .DemeGraph
    """
    with open(filename) as f:
        return loads(f.read())


def dumps(deme_graph):
    """
    Return a yaml-formatted string of the specified deme graph.

    :param .DemeGraph deme_graph: The deme graph to dump.
    :return: A yaml-formatted string.
    :rtype: str
    """
    d = deme_graph.asdict_compact()
    doc = as_document(d, schema=_deme_graph_schema)
    return doc.as_yaml()


def dump(deme_graph, filename):
    """
    Dump the specified deme graph to a yaml-formatted file.

    :param .DemeGraph deme_graph: The deme graph to dump.
    :param str filename: Name of the output file.
    """
    with open(filename, "w") as f:
        f.write(dumps(deme_graph))



import math
import unittest

import numpy as np
import moments

import demes.examples
from demes.convert import moments_


def moments_ooa(ns):
    fs = moments.Demographics1D.snm([sum(ns)])
    fs.integrate([1.6849315068493151], 0.2191780821917808)
    fs = moments.Manips.split_1D_to_2D(fs, ns[0], ns[1] + ns[2])
    fs.integrate(
        [1.6849315068493151, 0.2876712328767123],
        0.3254794520547945,
        m=[[0, 3.65], [3.65, 0]],
    )
    fs = moments.Manips.split_2D_to_3D_2(fs, ns[1], ns[2])

    def nu_func(t):
        return [
            1.6849315068493151,
            0.136986301369863
            * np.exp(
                np.log(4.071917808219178 / 0.136986301369863) * t / 0.05808219178082192
            ),
            0.06986301369863014
            * np.exp(
                np.log(7.409589041095891 / 0.06986301369863014)
                * t
                / 0.05808219178082192
            ),
        ]

    fs.integrate(
        nu_func,
        0.05808219178082192,
        m=[[0, 0.438, 0.2774], [0.438, 0, 1.4016], [0.2774, 1.4016, 0]],
    )
    return fs


class TestMomentsSFS(unittest.TestCase):
    # test function operations
    def test_convert_to_generations(self):
        g = demes.examples.gutenkunst_ooa()
        sample_times = [10, 20, 50]
        sample_times_gens = [s / g.generation_time for s in sample_times]
        g, sample_times = moments_.convert_to_generations(g, sample_times)
        self.assertTrue(
            np.all([s1 == s2 for s1, s2 in zip(sample_times, sample_times_gens)])
        )

    def test_num_lineages(self):
        # simple merge model
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="anc", initial_size=100, end_time=100)
        g.deme(id="pop1", initial_size=100, ancestors=["anc"], end_time=10)
        g.deme(id="pop2", initial_size=100, ancestors=["anc"], end_time=10)
        g.deme(id="pop3", initial_size=100, ancestors=["anc"], end_time=10)
        g.deme(
            id="pop",
            initial_size=100,
            ancestors=["pop1", "pop2", "pop3"],
            proportions=[0.1, 0.2, 0.7],
            start_time=10,
            end_time=0,
        )
        g.get_demographic_events()
        sampled_demes = ["pop"]
        demo_events, demes_present = moments_.get_demographic_events(g, sampled_demes)
        deme_sample_sizes = moments_.get_deme_sample_sizes(
            g, demo_events, sampled_demes, [20], demes_present
        )
        self.assertTrue(deme_sample_sizes[(math.inf, 100)][0] == 60)
        self.assertTrue(
            np.all([deme_sample_sizes[(100, 10)][i] == 20 for i in range(3)])
        )

        # simple admix model
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="anc", initial_size=100, end_time=100)
        g.deme(id="pop1", initial_size=100, ancestors=["anc"])
        g.deme(id="pop2", initial_size=100, ancestors=["anc"])
        g.deme(id="pop3", initial_size=100, ancestors=["anc"])
        g.deme(
            id="pop",
            initial_size=100,
            ancestors=["pop1", "pop2", "pop3"],
            proportions=[0.1, 0.2, 0.7],
            start_time=10,
            end_time=0,
        )
        g.get_demographic_events()
        sampled_demes = ["pop"]
        demo_events, demes_present = moments_.get_demographic_events(g, sampled_demes)
        deme_sample_sizes = moments_.get_deme_sample_sizes(
            g, demo_events, sampled_demes, [20], demes_present, unsampled_n=10
        )
        self.assertTrue(deme_sample_sizes[(math.inf, 100)][0] == 90)
        self.assertTrue(
            np.all([deme_sample_sizes[(100, 10)][i] == 30 for i in range(3)])
        )

    # test basic results against moments implementation
    def test_one_pop(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="Pop", initial_size=1000)
        fs = moments_.SFS(g, ["Pop"], [20])
        fs_m = moments.Demographics1D.snm([20])
        self.assertTrue(np.allclose(fs.data, fs_m.data))

        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="Pop", initial_size=1000, end_time=2000)
        g["Pop"].add_epoch(demes.Epoch(start_time=2000, end_time=0, initial_size=10000))
        fs = moments_.SFS(g, ["Pop"], [20])
        fs_m = moments.Demographics1D.snm([20])
        fs_m.integrate([10], 1)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_more_than_5_demes(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="anc", initial_size=1000, end_time=1000)
        for i in range(6):
            g.deme(id=f"pop{i}", initial_size=1000, ancestors=["anc"])
        with self.assertRaises(ValueError):
            moments_.SFS(g, ["pop{i}" for i in range(6)], [10 for i in range(6)])

        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="anc", initial_size=1000, end_time=1000)
        for i in range(3):
            g.deme(id=f"pop{i}", initial_size=1000, ancestors=["anc"])
        with self.assertRaises(ValueError):
            moments_.SFS(
                g,
                ["pop{i}" for i in range(3)],
                [10 for i in range(3)],
                sample_times=[5, 10, 15],
            )

    def test_one_pop_ancient_samples(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="Pop", initial_size=1000)
        fs = moments_.SFS(g, ["Pop", "Pop"], [20, 4], sample_times=[0, 100])
        fs_m = moments.Demographics1D.snm([24])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 20, 4)
        fs_m.integrate([1, 1], 100 / 2 / 1000, frozen=[False, True])
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_OOA(self):
        # this tests symmetric migration, size changes, splits
        g = demes.examples.gutenkunst_ooa()

        sample_sizes = [10, 10, 10]
        sampled_demes = ["YRI", "CEU", "CHB"]
        fs = moments_.SFS(g, sampled_demes, sample_sizes)
        # integrate with moments
        fs_m = moments_ooa(sample_sizes)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_simple_merge(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="Anc", initial_size=1000, end_time=100)
        g.deme(id="Source1", initial_size=2000, ancestors=["Anc"], end_time=10)
        g.deme(id="Source2", initial_size=3000, ancestors=["Anc"], end_time=10)
        g.deme(
            id="Pop",
            initial_size=4000,
            ancestors=["Source1", "Source2"],
            proportions=[0.8, 0.2],
            start_time=10,
        )
        g.get_demographic_events()
        fs = moments_.SFS(g, ["Pop"], [20])

        fs_m = moments.Demographics1D.snm([40])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 20, 20)
        fs_m.integrate([2, 3], 90 / 2 / 1000)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 20, 0.8)
        fs_m.integrate([4], 10 / 2 / 1000)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_simple_admixture(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="Anc", initial_size=1000, end_time=100)
        g.deme(id="Source1", initial_size=2000, ancestors=["Anc"])
        g.deme(id="Source2", initial_size=3000, ancestors=["Anc"])
        g.deme(
            id="Pop",
            initial_size=4000,
            ancestors=["Source1", "Source2"],
            proportions=[0.8, 0.2],
            start_time=10,
        )
        g.get_demographic_events()
        fs = moments_.SFS(g, ["Source1", "Source2", "Pop"], [10, 10, 10])

        fs_m = moments.Demographics1D.snm([40])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 20, 20)
        fs_m.integrate([2, 3], 90 / 2 / 1000)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 10, 0.8)
        fs_m.integrate([2, 3, 4], 10 / 2 / 1000)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_simple_growth_models(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(
            id="Pop",
            epochs=[
                demes.Epoch(end_time=1000, initial_size=1000),
                demes.Epoch(initial_size=500, final_size=5000, end_time=0),
            ],
        )
        fs = moments_.SFS(g, ["Pop"], [100])

        fs_m = moments.Demographics1D.snm([100])

        def nu_func(t):
            return [0.5 * np.exp(np.log(5000 / 500) * t / 0.5)]

        fs_m.integrate(nu_func, 0.5)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(
            id="Pop",
            epochs=[
                demes.Epoch(end_time=1000, initial_size=1000),
                demes.Epoch(
                    initial_size=500,
                    final_size=5000,
                    end_time=0,
                    size_function="linear",
                ),
            ],
        )
        fs = moments_.SFS(g, ["Pop"], [100])

        fs_m = moments.Demographics1D.snm([100])

        def nu_func(t):
            return [0.5 + t / 0.5 * (5 - 0.5)]

        fs_m.integrate(nu_func, 0.5)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_simple_pulse_model(self):
        g = demes.DemeGraph(description="test", time_units="generations")
        g.deme(id="anc", initial_size=1000, end_time=100)
        g.deme(id="source", initial_size=1000, ancestors=["anc"])
        g.deme(id="dest", initial_size=1000, ancestors=["anc"])
        g.pulse(source="source", dest="dest", time=10, proportion=0.1)
        g.get_demographic_events()
        fs = moments_.SFS(g, ["source", "dest"], [20, 20])

        fs_m = moments.Demographics1D.snm([60])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 40, 20)
        fs_m.integrate([1, 1], 90 / 2 / 1000)
        fs_m = moments.Manips.admix_inplace(fs_m, 0, 1, 20, 0.1)
        fs_m.integrate([1, 1], 10 / 2 / 1000)
        self.assertTrue(np.allclose(fs.data, fs_m.data))

    def test_n_way_split(self):
        g = demes.DemeGraph(description="three-way", time_units="generations")
        g.deme(id="anc", initial_size=1000, end_time=10)
        g.deme(id="deme1", initial_size=1000, ancestors=["anc"])
        g.deme(id="deme2", initial_size=1000, ancestors=["anc"])
        g.deme(id="deme3", initial_size=1000, ancestors=["anc"])
        g.get_demographic_events()
        ns = [10, 15, 20]
        fs = moments_.SFS(g, ["deme1", "deme2", "deme3"], ns)
        self.assertTrue(np.all([fs.sample_sizes[i] == ns[i] for i in range(len(ns))]))

        fs_m1 = moments.Demographics1D.snm([sum(ns)])
        fs_m1 = moments.Manips.split_1D_to_2D(fs_m1, ns[0], ns[1] + ns[2])
        fs_m1 = moments.Manips.split_2D_to_3D_2(fs_m1, ns[1], ns[2])
        fs_m1.integrate([1, 1, 1], 10 / 2 / 1000)

        fs_m2 = moments.Demographics1D.snm([sum(ns)])
        fs_m2 = moments.Manips.split_1D_to_2D(fs_m2, ns[0] + ns[1], ns[2])
        fs_m2 = moments.Manips.split_2D_to_3D_1(fs_m2, ns[0], ns[1])
        fs_m2 = fs_m2.swapaxes(1, 2)
        fs_m2.integrate([1, 1, 1], 10 / 2 / 1000)

        self.assertTrue(np.allclose(fs.data, fs_m1.data))
        self.assertTrue(np.allclose(fs.data, fs_m2.data))

    def test_n_way_admixture(self):
        g = demes.DemeGraph(description="three-way merge", time_units="generations")
        g.deme(id="anc", initial_size=1000, end_time=100)
        g.deme(id="source1", initial_size=1000, end_time=10, ancestors=["anc"])
        g.deme(id="source2", initial_size=1000, end_time=10, ancestors=["anc"])
        g.deme(id="source3", initial_size=1000, end_time=10, ancestors=["anc"])
        g.deme(
            id="merged",
            initial_size=1000,
            ancestors=["source1", "source2", "source3"],
            proportions=[0.5, 0.2, 0.3],
            start_time=10,
        )
        g.get_demographic_events()
        ns = [10]
        fs = moments_.SFS(g, ["merged"], ns)

        fs_m = moments.Demographics1D.snm([30])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 10, 20)
        fs_m = moments.Manips.split_2D_to_3D_2(fs_m, 10, 10)
        fs_m.integrate([1, 1, 1], 90 / 2 / 1000)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 10, 0.5 / 0.7)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 10, 0.3)
        fs_m.integrate([1], 10 / 2 / 1000)

        self.assertTrue(np.allclose(fs_m.data, fs.data))

        g = demes.DemeGraph(description="three-way admix", time_units="generations")
        g.deme(id="anc", initial_size=1000, end_time=100)
        g.deme(id="source1", initial_size=1000, ancestors=["anc"])
        g.deme(id="source2", initial_size=1000, ancestors=["anc"])
        g.deme(id="source3", initial_size=1000, ancestors=["anc"])
        g.deme(
            id="admixed",
            initial_size=1000,
            ancestors=["source1", "source2", "source3"],
            proportions=[0.5, 0.2, 0.3],
            start_time=10,
        )
        g.get_demographic_events()
        ns = [10]
        fs = moments_.SFS(g, ["admixed"], ns)

        fs_m = moments.Demographics1D.snm([30])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 10, 20)
        fs_m = moments.Manips.split_2D_to_3D_2(fs_m, 10, 10)
        fs_m.integrate([1, 1, 1], 90 / 2 / 1000)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 10, 0.5 / 0.7)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 10, 0.3)
        fs_m.integrate([1], 10 / 2 / 1000)

        self.assertTrue(np.allclose(fs_m.data[1:-1], fs.data[1:-1]))

        fs = moments_.SFS(g, ["source1", "admixed"], [10, 10])

        fs_m = moments.Demographics1D.snm([40])
        fs_m = moments.Manips.split_1D_to_2D(fs_m, 20, 20)
        fs_m = moments.Manips.split_2D_to_3D_2(fs_m, 10, 10)
        fs_m.integrate([1, 1, 1], 90 / 2 / 1000)
        fs_m = moments.Manips.admix_into_new(fs_m, 0, 1, 10, 0.5 / 0.7)
        fs_m = moments.Manips.admix_into_new(fs_m, 1, 2, 10, 0.3)
        fs_m.integrate([1, 1], 10 / 2 / 1000)

        fs[0, 0] = fs[-1, -1] = 0
        fs_m[0, 0] = fs_m[-1, -1] = 0
        self.assertTrue(np.allclose(fs_m.data[1:-1], fs.data[1:-1]))



import demes

g1 = demes.load("gutenkunst_ooa.yml")
demes.dump(g1, "gutenkunst_ooa_output.yml")

g2 = demes.load("browning_america.yml")
demes.dump(g2, "browning_america_output.yml")

# an example with overlapping ancestors/descendants, mass migration event,
# symmetric migration, and asymmetric migration
g3 = demes.load("offshoots.yml")
demes.dump(g3, "offshoots_output.yml")



import unittest

import demes
import demes.examples


class TestExamples(unittest.TestCase):
    maxDiff = None

    def dump_and_load(self, g):
        s = demes.dumps(g)
        g2 = demes.loads(s)
        s2 = demes.dumps(g2)
        g3 = demes.loads(s2)
        self.assertEqual(g2, g3)
        self.assertEqual(s, s2)

    def test_zigzag(self):
        self.dump_and_load(demes.examples.zigzag())

    def test_gutenkunst_ooa(self):
        self.dump_and_load(demes.examples.gutenkunst_ooa())

    def test_browning_america(self):
        self.dump_and_load(demes.examples.browning_america())

    def test_jacobs_papuans(self):
        self.dump_and_load(demes.examples.jacobs_papuans())

    def test_IM(self):
        self.dump_and_load(demes.examples.IM(10, 10000, 1e-5))



import typing
from typing import List, Mapping, Union, Optional
import itertools
import math
import collections
import copy
import operator

import attr
from attr.validators import optional

from .script import dumps, loads

Number = Union[int, float]
ID = str
Time = Number
Size = Number
Rate = float
Proportion = float

_ISCLOSE_REL_TOL = 1e-9
_ISCLOSE_ABS_TOL = 1e-12

# Validator functions.


def positive(self, attribute, value):
    if value <= 0:
        raise ValueError(f"{attribute.name} must be greater than zero")


def non_negative(self, attribute, value):
    if value < 0:
        raise ValueError(f"{attribute.name} must be non-negative")


def finite(self, attribute, value):
    if math.isinf(value):
        raise ValueError(f"{attribute.name} must be finite")


def unit_interval(self, attribute, value):
    if not (0 <= value <= 1):
        raise ValueError(f"must have 0 <= {attribute.name} <= 1")


def isclose(
    a: Optional[Number],
    b: Optional[Number],
    *,
    rel_tol=_ISCLOSE_REL_TOL,
    abs_tol=_ISCLOSE_ABS_TOL,
) -> bool:
    """
    Wrapper around math.isclose() that handles None.
    """
    if None in (a, b):
        return (a,) == (b,)
    else:
        return math.isclose(
            typing.cast(float, a),
            typing.cast(float, b),
            rel_tol=rel_tol,
            abs_tol=abs_tol,
        )


def isclose_deme_proportions(
    a_ids: Optional[List[ID]],
    a_proportions: Optional[List[Proportion]],
    b_ids: Optional[List[ID]],
    b_proportions: Optional[List[Proportion]],
    *,
    rel_tol=_ISCLOSE_REL_TOL,
    abs_tol=_ISCLOSE_ABS_TOL,
) -> bool:
    """
    Returns true if (a_ids, a_proportions) and (b_ids, b_proportions)
    are semantically equivalent. The order of ids is ignored, and proportions
    are checked for numerical closeness.
    """
    if None in (a_ids, b_ids):
        return (a_ids, a_proportions) == (b_ids, b_proportions)

    a_ids = typing.cast(List[ID], a_ids)
    a_proportions = typing.cast(List[Proportion], a_proportions)
    b_ids = typing.cast(List[ID], b_ids)
    b_proportions = typing.cast(List[Number], b_proportions)

    if len(a_ids) != len(b_ids) or len(a_proportions) != len(b_proportions):
        return False
    a = sorted(zip(a_ids, a_proportions), key=operator.itemgetter(0))
    b = sorted(zip(b_ids, b_proportions), key=operator.itemgetter(0))
    for (a_id, a_proportion), (b_id, b_proportion) in zip(a, b):
        if a_id != b_id or not isclose(
            a_proportion, b_proportion, rel_tol=rel_tol, abs_tol=abs_tol
        ):
            return False
    return True


@attr.s(auto_attribs=True)
class Epoch:
    """
    Population size parameters for a deme in a specified time period.
    Times follow the forwards-in-time convention (time values increase
    from the present towards the past). The start time of the epoch is
    the more ancient time, and the end time is more recent, so that the
    start time must be greater than the end time

    :ivar start_time: The start time of the epoch.
    :ivar end_time: The end time of the epoch (must be specified).
    :ivar initial_size: Population size at ``start_time``.
    :ivar final_size: Population size at ``end_time``.
        If ``initial_size != final_size``, the population size changes
        monotonically between the start and end times.
    :ivar size_function: The size change function. Common options are constant,
        exponential, or linear, though any string is valid. Warning: downstream
        simulators might not understand the size_function provided.
    :ivar selfing_rate: An optional selfing rate for this epoch.
    :ivar cloning_rate: An optional cloning rate for this epoch.
    """

    start_time: Optional[Time] = attr.ib(default=None, validator=optional(non_negative))
    end_time: Time = attr.ib(default=None, validator=[non_negative, finite])
    initial_size: Optional[Size] = attr.ib(
        default=None, validator=optional([positive, finite])
    )
    final_size: Optional[Size] = attr.ib(
        default=None, validator=optional([positive, finite])
    )
    size_function: Optional[str] = attr.ib(default=None)
    selfing_rate: Optional[Proportion] = attr.ib(
        default=None, validator=optional(unit_interval)
    )
    cloning_rate: Optional[Proportion] = attr.ib(
        default=None, validator=optional(unit_interval)
    )

    def __attrs_post_init__(self):
        if self.initial_size is None and self.final_size is None:
            raise ValueError("must set either initial_size or final_size")
        if (
            self.start_time is not None
            and self.end_time is not None
            and self.start_time <= self.end_time
        ):
            raise ValueError("must have start_time > end_time")
        if (
            self.start_time is not None
            and self.initial_size is not None
            and self.final_size is not None
        ):
            if math.isinf(self.start_time) and self.initial_size != self.final_size:
                raise ValueError("if start time is inf, must be a constant size epoch")

    @property
    def time_span(self):
        """
        The time span of the epoch.
        """
        return self.start_time - self.end_time

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and isclose(
                self.start_time, other.start_time, rel_tol=rel_tol, abs_tol=abs_tol
            )
            and isclose(self.end_time, other.end_time, rel_tol=rel_tol, abs_tol=abs_tol)
            and isclose(
                self.initial_size, other.initial_size, rel_tol=rel_tol, abs_tol=abs_tol
            )
            and isclose(
                self.final_size, other.final_size, rel_tol=rel_tol, abs_tol=abs_tol
            )
            and self.size_function == other.size_function
            and isclose(
                self.selfing_rate, other.selfing_rate, rel_tol=rel_tol, abs_tol=abs_tol
            )
            and isclose(
                self.cloning_rate, other.cloning_rate, rel_tol=rel_tol, abs_tol=abs_tol
            )
        )


@attr.s(auto_attribs=True)
class Migration:
    """
    Parameters for continuous migration from one deme to another.
    Source and destination demes follow the forwards-in-time convention,
    of migrations born in the source deme having children in the dest
    deme.

    :ivar source: The source deme.
    :ivar dest: The destination deme.
    :ivar start_time: The time at which the migration rate becomes activate.
    :ivar end_time: The time at which the migration rate is deactivated.
    :ivar rate: The rate of migration. Set to zero to disable migrations after
        the given time.
    """

    source: ID = attr.ib()
    dest: ID = attr.ib()
    start_time: Time = attr.ib(validator=non_negative)
    end_time: Time = attr.ib(validator=[non_negative, finite])
    rate: Rate = attr.ib(validator=[non_negative, finite])

    def __attrs_post_init__(self):
        if self.source == self.dest:
            raise ValueError("source and dest cannot be the same deme")

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and self.source == other.source
            and self.dest == other.dest
            and isclose(
                self.start_time, other.start_time, rel_tol=rel_tol, abs_tol=abs_tol
            )
            and isclose(self.end_time, other.end_time, rel_tol=rel_tol, abs_tol=abs_tol)
            and isclose(self.rate, other.rate, rel_tol=rel_tol, abs_tol=abs_tol)
        )


@attr.s(auto_attribs=True)
class Pulse:
    """
    Parameters for a pulse of migration from one deme to another.
    Source and destination demes follow the forwards-in-time convention,
    of migrations born in the source deme having children in the dest
    deme.

    :ivar source: The source deme.
    :ivar dest: The destination deme.
    :ivar time: The time of migration.
    :ivar proportion: At the instant after migration, this is the proportion
        of individuals in the destination deme made up of individuals from
        the source deme.
    """

    source: ID = attr.ib()
    dest: ID = attr.ib()
    time: Time = attr.ib(validator=[non_negative, finite])
    proportion: Proportion = attr.ib(validator=unit_interval)

    def __attrs_post_init__(self):
        if self.source == self.dest:
            raise ValueError("source and dest cannot be the same deme")

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and self.source == other.source
            and self.dest == other.dest
            and isclose(self.time, other.time, rel_tol=rel_tol, abs_tol=abs_tol)
            and isclose(
                self.proportion, other.proportion, rel_tol=rel_tol, abs_tol=abs_tol
            )
        )


@attr.s(auto_attribs=True)
class Split:
    """
    Parameters for a split event, in which a deme ends at a given time and
    contributes ancestry to an arbitrary number of descendant demes. Note
    that there could be just a single descendant deme, in which case ``split``
    is a bit of a misnomer...

    :ivar parent: The parental deme.
    :ivar children: A list of descendant demes.
    :ivar time: The split time.
    """

    parent: ID = attr.ib()
    children: List[ID] = attr.ib()
    time: Time = attr.ib(validator=[non_negative, finite])

    def __attrs_post_init__(self):
        if not isinstance(self.children, list):
            raise ValueError("children of split must be passed as a list")
        for child in self.children:
            if child == self.parent:
                raise ValueError("child and parent cannot be the same deme")

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and self.parent == other.parent
            and sorted(self.children) == sorted(other.children)
            and isclose(self.time, other.time, rel_tol=rel_tol, abs_tol=abs_tol)
        )


@attr.s(auto_attribs=True)
class Branch:
    """
    Parameters for a branch event, where a new deme branches off from a parental
    deme. The parental deme need not end at that time.

    :ivar parent: The parental deme.
    :ivar child: The descendant deme.
    :ivar time: The branch time.
    """

    parent: ID = attr.ib()
    child: ID = attr.ib()
    time: Time = attr.ib(validator=[non_negative, finite])

    def __attrs_post_init__(self):
        if self.child == self.parent:
            raise ValueError("child and parent cannot be the same deme")

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and self.parent == other.parent
            and self.child == other.child
            and isclose(self.time, other.time, rel_tol=rel_tol, abs_tol=abs_tol)
        )


@attr.s(auto_attribs=True)
class Merge:
    """
    Parameters for a merge event, in which two or more demes end at some time and
    contribute to a descendant deme.

    :ivar parents: A list of parental demes.
    :ivar proportions: A list of ancestry proportions, in order of `parents`.
    :ivar child: The descendant deme.
    :ivar time: The merge time.
    """

    parents: List[ID] = attr.ib()
    proportions: List[Proportion] = attr.ib()
    child: ID = attr.ib()
    time: Time = attr.ib(validator=[non_negative, finite])

    def __attrs_post_init__(self):
        if not isinstance(self.parents, list):
            raise ValueError("parents must be passed as a list")
        if not isinstance(self.proportions, list):
            raise ValueError("proportions must be passed as a list")
        if len(self.parents) < 2:
            raise ValueError("merge must involve at least two ancestors")
        if math.isclose(sum(self.proportions), 1) is False:
            raise ValueError("proportions must sum to 1")
        if len(self.parents) != len(self.proportions):
            raise ValueError("parents and proportions must have same length")
        if self.child in self.parents:
            raise ValueError("merged deme cannot be its own ancestor")
        if len(set(self.parents)) != len(self.parents):
            raise ValueError("cannot repeat parents in merge")

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and isclose_deme_proportions(
                self.parents,
                self.proportions,
                other.parents,
                other.proportions,
                rel_tol=rel_tol,
                abs_tol=abs_tol,
            )
            and self.child == other.child
            and isclose(self.time, other.time, rel_tol=rel_tol, abs_tol=abs_tol)
        )


@attr.s(auto_attribs=True)
class Admix:
    """
    Parameters for an admixture event, where two or more demes contribute ancestry
    to a new deme.

    :ivar parents: A list of source demes.
    :ivar proportions: A list of ancestry proportions, in order of `parents`.
    :ivar child: The admixed deme.
    :ivar time: The admixture time.
    """

    parents: List[ID] = attr.ib()
    proportions: List[Proportion] = attr.ib()
    child: ID = attr.ib()
    time: Time = attr.ib(validator=[non_negative, finite])

    def __attrs_post_init__(self):
        if not isinstance(self.parents, list):
            raise ValueError("parents must be passed as a list")
        if not isinstance(self.proportions, list):
            raise ValueError("proportions must be passed as a list")
        if len(self.parents) < 2:
            raise ValueError("admixture must involve at least two ancestors")
        if math.isclose(sum(self.proportions), 1) is False:
            raise ValueError("Proportions must sum to 1")
        if len(self.parents) != len(self.proportions):
            raise ValueError("parents and proportions must have same length")
        if self.child in self.parents:
            raise ValueError("admixed deme cannot be its own ancestor")
        if len(set(self.parents)) != len(self.parents):
            raise ValueError("cannot repeat parents in admixure")

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and isclose_deme_proportions(
                self.parents,
                self.proportions,
                other.parents,
                other.proportions,
                rel_tol=rel_tol,
                abs_tol=abs_tol,
            )
            and self.child == other.child
            and isclose(self.time, other.time, rel_tol=rel_tol, abs_tol=abs_tol)
        )


@attr.s(auto_attribs=True)
class Deme:
    """
    A collection of individuals that are exchangeable at any fixed time.

    :ivar id: A string identifier for the deme.
    :ivar description: An optional description of the deme.
    :ivar ancestors: List of ancestors to the deme.
        If the deme has no ancestors, this should be ``None``, and cannot
        be used with ``ancestor``.
    :ivar proportions: List of proportions if ``ancestors`` is given.
    :ivar epochs: A list of epochs, which define the population size(s) of
        the deme. The deme must be initially created with exactly one epoch.
        Additional epochs may be added with :meth:`.add_epoch`
    :vartype epochs: list of :class:`.Epoch`
    :ivar selfing_rate: An optional selfing rate for this deme.
    :ivar cloning_rate: An optional cloning rate for this deme.
    """

    id: ID = attr.ib()
    description: str = attr.ib()
    ancestors: List[ID] = attr.ib()
    proportions: List[Proportion] = attr.ib()
    epochs: List[Epoch] = attr.ib()
    selfing_rate: Optional[Proportion] = attr.ib(
        default=None, validator=optional([unit_interval])
    )
    cloning_rate: Optional[Proportion] = attr.ib(
        default=None, validator=optional([unit_interval])
    )

    @epochs.validator
    def _check_epochs(self, attribute, value):
        if len(self.epochs) != 1:
            raise ValueError(
                "Deme must be created with exactly one epoch."
                "Use add_epoch() to supply additional epochs."
            )

    def __attrs_post_init__(self):
        if self.ancestors is not None:
            if not isinstance(self.ancestors, (list, tuple)):
                raise TypeError("ancestors must be a list of deme IDs")
            if len(self.ancestors) > 1 and self.proportions is None:
                raise ValueError("proportions must be set if more than one ancestor")
            if len(self.ancestors) != len(self.proportions):
                raise ValueError("ancestors and proportions must have same length")
            if self.id in self.ancestors:
                raise ValueError(f"{self.id} cannot be its own ancestor")
        # if selfing or cloning rates are not given, set them to deme's default rate
        epoch = self.epochs[0]
        if epoch.selfing_rate is None:
            epoch.selfing_rate = self.selfing_rate
        if epoch.cloning_rate is None:
            epoch.cloning_rate = self.cloning_rate

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        return (
            self.__class__ is other.__class__
            and self.id == other.id
            and isclose_deme_proportions(
                self.ancestors,
                self.proportions,
                other.ancestors,
                other.proportions,
                rel_tol=rel_tol,
                abs_tol=abs_tol,
            )
            and all(
                e1.isclose(e2, rel_tol=rel_tol, abs_tol=abs_tol)
                for e1, e2 in zip(self.epochs, other.epochs)
            )
        )

    def add_epoch(self, epoch: Epoch):
        """
        Add an epoch to the deme's epoch list.
        Epochs must be non overlapping and added in time-decreasing order, i.e.
        starting with the most ancient epoch and adding epochs sequentially toward
        present.

        :param .Epoch epoch: The epoch to add.
        """
        assert len(self.epochs) > 0
        # if the epoch start time is not given, it equals the previous epoch's end time
        prev_epoch = self.epochs[-1]
        if epoch.start_time is None:
            epoch.start_time = prev_epoch.end_time
        elif epoch.start_time > prev_epoch.end_time:
            raise ValueError(
                "epochs must be non overlapping and added in time-decreasing order"
            )
        if prev_epoch.end_time != epoch.start_time:
            raise ValueError("cannot have gap between consecutive epochs")
        if epoch.time_span <= 0:
            raise ValueError("epoch must exist for some positive time")
        # implicitly set the initial and final sizes, if not given
        if epoch.initial_size is None:
            epoch.initial_size = prev_epoch.final_size
        if epoch.final_size is None:
            epoch.final_size = epoch.initial_size
        # check or assign the size function over this epoch
        if epoch.size_function is None:
            if epoch.initial_size == epoch.final_size:
                epoch.size_function = "constant"
            else:
                epoch.size_function = "exponential"
        else:
            # check if constant function is correct
            if (
                epoch.size_function == "constant"
                and epoch.initial_size != epoch.final_size
            ):
                raise ValueError(
                    "epoch size function is constant but initial and "
                    "final sizes are not equal"
                )
        # if selfing or cloning rates are not given, set them to deme's default rate
        if epoch.selfing_rate is None:
            epoch.selfing_rate = self.selfing_rate
        if epoch.cloning_rate is None:
            epoch.cloning_rate = self.cloning_rate
        self.epochs.append(epoch)

    @property
    def start_time(self):
        """
        The start time of the deme's existence.
        """
        return self.epochs[0].start_time

    @property
    def end_time(self):
        """
        The end time of the deme's existence.
        """
        return self.epochs[-1].end_time

    @property
    def time_span(self):
        """
        The time span over which the deme exists.
        """
        return self.start_time - self.end_time


@attr.s(auto_attribs=True)
class DemeGraph:
    """
    A directed graph that describes a demography. Vertices are demes and edges
    correspond to ancestor/descendent relations. Edges are directed from
    ancestors to descendants.

    :ivar description: A human readable description of the demography.
    :ivar time_units: The units of time used for the demography. This is
        commonly ``years`` or ``generations``, but can be any string.
        This field is intended to be useful for documenting a demography,
        but the actual value provided here should not be relied upon.
    :ivar generation_time: The generation time of demes, in units given
        by the ``time_units`` parameter. Concretely, dividing all times
        by ``generation_time`` will convert the deme graph to have time
        units in generations.  If ``generation_time`` is ``None``, the units
        are assumed to be in generations already.
        See also: :meth:`.in_generations`.
    :ivar default_Ne: The default population size to use when creating new
        demes with :meth:`.deme`. May be ``None``.
    :ivar doi: If the deme graph describes a published demography, the DOI
        should be be given here. May be ``None``.
    :ivar demes: A list of demes in the demography.
        Not intended to be passed when the deme graph is instantiated.
        Use :meth:`.deme` instead.
    :vartype demes: list of :class:`.Deme`
    :ivar migrations: A list of continuous migrations for the demography.
        Not intended to be passed when the deme graph is instantiated.
        Use :meth:`migration` or :meth:`symmetric_migration` instead.
    :vartype migrations: list of :class:`.Migration`
    :ivar pulses: A list of migration pulses for the demography.
        Not intended to be passed when the deme graph is instantiated.
        Use :meth:`pulse` instead.
    """

    description: str = attr.ib()
    time_units: str = attr.ib()
    generation_time: Optional[Time] = attr.ib(
        default=None, validator=optional([positive, finite])
    )
    default_Ne: Optional[Size] = attr.ib(
        default=None, validator=optional([positive, finite])
    )
    doi: Optional[str] = attr.ib(default=None)
    demes: List[Deme] = attr.ib(factory=list)
    migrations: List[Migration] = attr.ib(factory=list)
    pulses: List[Pulse] = attr.ib(factory=list)
    splits: List[Split] = attr.ib(factory=list)
    branches: List[Branch] = attr.ib(factory=list)
    mergers: List[Merge] = attr.ib(factory=list)
    admixtures: List[Admix] = attr.ib(factory=list)
    selfing_rate: Proportion = attr.ib(default=None)
    cloning_rate: Proportion = attr.ib(default=None)

    def __attrs_post_init__(self):
        self._deme_map: Mapping[ID, Deme] = dict()

    def __getitem__(self, deme_id):
        """
        Return the :class:`.Deme` with the specified id.
        """
        return self._deme_map[deme_id]

    def __contains__(self, deme_id):
        """
        Check if the deme graph contains a deme with the specified id.
        """
        return deme_id in self._deme_map

    def isclose(
        self,
        other,
        *,
        rel_tol=_ISCLOSE_REL_TOL,
        abs_tol=_ISCLOSE_ABS_TOL,
    ) -> bool:
        def sorted_eq(aa, bb, *, rel_tol, abs_tol) -> bool:
            # Order-agnostic equality check.
            if len(aa) != len(bb):
                return False
            for (a, b) in zip(sorted(aa), sorted(bb)):
                if not a.isclose(b, rel_tol=rel_tol, abs_tol=abs_tol):
                    return False
            return True

        return (
            self.__class__ is other.__class__
            and self.time_units == other.time_units
            and self.generation_time == other.generation_time
            and sorted_eq(self.demes, other.demes, rel_tol=rel_tol, abs_tol=abs_tol)
            and sorted_eq(
                self.migrations, other.migrations, rel_tol=rel_tol, abs_tol=abs_tol
            )
            and sorted_eq(self.pulses, other.pulses, rel_tol=rel_tol, abs_tol=abs_tol)
        )

    def deme(
        self,
        id,
        description=None,
        ancestors=None,
        proportions=None,
        start_time=None,
        end_time=None,
        initial_size=None,
        final_size=None,
        epochs=None,
        selfing_rate=None,
        cloning_rate=None,
    ):
        """
        Add a deme to the graph.

        :param str id: A string identifier for the deme.
        :param list ancestors: A list of ancestors of this deme. May be ``None``.
            If ``len(ancestors) > 1``, must also give ``proportions``.
        :param list proportions: A list of proportions of ancestory for ``ancestors``.
            Proportions must sum to 1.
        :param start_time: The time at which this deme begins existing.
        :param end_time: The time at which this deme stops existing.
            If the deme has an ancestor the ``end_time`` will be set to the
            ancestor's ``start_time``.
        :param initial_size: The initial population size of the deme. If ``None``,
            this is taken from the deme graph's ``default_Ne`` field.
        :param final_size: The final population size of the deme. If ``None``,
            the deme has a constant ``initial_size`` population size.
        :param epochs: Additional epochs that define population size changes for
            the deme.
        :ivar selfing_rate: An optional selfing rate for this deme.
        :ivar cloning_rate: An optional cloning rate for this deme.
        """
        if initial_size is None:
            initial_size = self.default_Ne
        if initial_size is None and epochs is not None:
            initial_size = epochs[0].initial_size
        if initial_size is None:
            raise ValueError(f"must set initial_size for {id}")
        if selfing_rate is None:
            selfing_rate = self.selfing_rate
        if cloning_rate is None:
            cloning_rate = self.cloning_rate
        # set the start time to inf or to the ancestors end times, if not given
        if ancestors is not None:
            if not isinstance(ancestors, (list, tuple)):
                raise TypeError("ancestors must be a list of deme IDs")
            if len(ancestors) > 1 and start_time is None:
                raise ValueError("must specify start time if more than one ancestor")
            if ancestors[0] in self and start_time is None:
                start_time = self[ancestors[0]].epochs[-1].end_time
            if len(ancestors) == 1 and proportions is None:
                proportions = [1.0]
        else:
            if start_time is None:
                start_time = float("inf")
        # build the deme, and then add epochs as necessary
        if epochs is None:
            # if epochs are not given, we assign a single epoch over that deme
            if final_size is None:
                final_size = initial_size
            if end_time is None:
                end_time = 0
            if initial_size == final_size:
                size_function = "constant"
            else:
                size_function = "exponential"
            epoch = Epoch(
                start_time,
                end_time,
                initial_size=initial_size,
                final_size=final_size,
                size_function=size_function,
                selfing_rate=selfing_rate,
                cloning_rate=cloning_rate,
            )
            deme = Deme(
                id,
                description,
                ancestors,
                proportions,
                [epoch],
                selfing_rate,
                cloning_rate,
            )
        else:
            if end_time is None:
                end_time = epochs[-1].end_time
            if end_time != epochs[-1].end_time:
                raise ValueError("deme and final epoch end times do not align")
            if epochs[0].selfing_rate is None:
                epochs[0].selfing_rate = selfing_rate
            if epochs[0].cloning_rate is None:
                epochs[0].cloning_rate = cloning_rate
            # deal with first epoch and deme start times
            if epochs[0].start_time is None:
                # first epoch starts at deme start time
                epochs[0].start_time = start_time
            elif epochs[0].start_time < start_time:
                # insert const size epoch to reach the start of first listed epoch
                epochs.insert(
                    0,
                    Epoch(
                        start_time,
                        epochs[0].start_time,
                        initial_size=initial_size,
                        final_size=initial_size,
                        size_function="constant",
                        selfing_rate=selfing_rate,
                        cloning_rate=cloning_rate,
                    ),
                )
            elif epochs[0].start_time > start_time:
                raise ValueError(
                    "first epoch start time must be less than or equal to "
                    "deme start time"
                )
            # set up sizes of first deme, since subsequent demes are built off of it
            if epochs[0].final_size is None:
                epochs[0].final_size = epochs[0].initial_size
            if epochs[0].size_function is None:
                if epochs[0].initial_size == epochs[0].final_size:
                    epochs[0].size_function = "constant"
                else:
                    epochs[0].size_function = "exponential"
            deme = Deme(
                id,
                description,
                ancestors,
                proportions,
                [epochs[0]],
                selfing_rate,
                cloning_rate,
            )
            for epoch in epochs[1:]:
                deme.add_epoch(epoch)
        self._deme_map[deme.id] = deme
        self.demes.append(deme)

    def check_time_intersection(self, deme1, deme2, time):
        deme1 = self[deme1]
        deme2 = self[deme2]
        time_lo = max(deme1.end_time, deme2.end_time)
        time_hi = min(deme1.start_time, deme2.start_time)
        if time is not None:
            if not (time_lo <= time <= time_hi):
                raise ValueError(
                    f"{time} not in interval [{time_lo}, {time_hi}], "
                    f"as defined by the time-intersection of {deme1.id} "
                    f"(start_time={deme1.start_time}, end_time={deme1.end_time}) "
                    f"and {deme2.id} (start_time={deme2.start_time}, "
                    f"end_time={deme2.end_time})."
                )
        return time_lo, time_hi

    def symmetric_migration(self, demes=[], rate=0, start_time=None, end_time=None):
        """
        Add continuous symmetric migrations between all pairs of demes in a list.

        :param demes: list of deme IDs. Migration is symmetric between all
            pairs of demes in this list.
        :param rate: The rate of migration per generation.
        :param start_time: The time at which the migration rate is enabled.
        :param end_time: The time at which the migration rate is disabled.
        """
        if len(demes) < 2:
            raise ValueError("must specify two or more demes")
        for source, dest in itertools.permutations(demes, 2):
            self.migration(source, dest, rate, start_time, end_time)

    def migration(self, source, dest, rate=0, start_time=None, end_time=None):
        """
        Add continuous migration from one deme to another.
        Source and destination demes follow the forwards-in-time convention,
        so that the migration rate refers to the movement of individuals from
        the ``source`` deme to the ``dest`` deme.

        :param source: The source deme.
        :param dest: The destination deme.
        :param rate: The rate of migration per generation.
        :param start_time: The time at which the migration rate is enabled.
            If ``None``, the start time is defined by the earliest time at
            which the demes coexist.
        :param end_time: The time at which the migration rate is disabled.
            If ``None``, the end time is defined by the latest time at which
            the demes coexist.
        """
        for deme_id in (source, dest):
            if deme_id not in self:
                raise ValueError(f"{deme_id} not in deme graph")
        time_lo, time_hi = self.check_time_intersection(source, dest, start_time)
        if start_time is None:
            start_time = time_hi
        else:
            self.check_time_intersection(source, dest, start_time)
        if end_time is None:
            end_time = time_lo
        else:
            self.check_time_intersection(source, dest, end_time)
        self.migrations.append(Migration(source, dest, start_time, end_time, rate))

    def pulse(self, source, dest, proportion, time):
        """
        Add a pulse of migration at a fixed time.
        Source and destination demes follow the forwards-in-time convention.

        :param source: The source deme.
        :param dest: The destination deme.
        :param proportion: At the instant after migration, this is the expected
            proportion of individuals in the destination deme made up of individuals
            from the source deme.
        :param time: The time at which migrations occur.
        """
        for deme_id in (source, dest):
            if deme_id not in self:
                raise ValueError(f"{deme_id} not in deme graph")
        self.check_time_intersection(source, dest, time)
        self.pulses.append(Pulse(source, dest, time, proportion))

    @property
    def successors(self):
        """
        Lists of successors for all demes in the graph.
        """
        # use collections.defaultdict(list)
        succ = {}
        for deme_info in self.demes:
            succ.setdefault(deme_info.id, [])
            if deme_info.ancestors is not None:
                for a in deme_info.ancestors:
                    succ.setdefault(a, [])
                    succ[a].append(deme_info.id)
        return succ

    @property
    def predecessors(self):
        """
        Lists of predecessors (ancestors) for all demes in the graph.
        """
        pred = {}
        for deme_info in self.demes:
            pred.setdefault(deme_info.id, [])
            if deme_info.ancestors is not None:
                for a in deme_info.ancestors:
                    pred[deme_info.id].append(a)
        return pred

    def split(self, parent, children, time):
        """
        Add split event at a given time. Split events involve a parental deme
        whose end time equals the start time of all children demes.

        :param parent: The ancestral deme.
        :param children: A list of descendant demes.
        :param time: The time at which split occurs.
        """
        for child in children:
            # check parent/children relationship and end/start times
            if child == parent:
                raise ValueError("cannot be ancestor of own deme")
            if self[parent].end_time != self[child].start_time:
                raise ValueError(
                    f"{parent} and {child} must have matching end and start times"
                )
            # the ancestor of each child population is set
            self[child].ancestors = [parent]
        self.splits.append(Split(parent, children, time))

    def branch(self, parent, child, time):
        """
        Add branch event at a given time.

        :param parent: The ancestral deme.
        :param children: A list of descendant demes.
        :param time: The time at which branch event occurs.
        """
        if (
            self[child].start_time < self[parent].end_time
            or self[child].start_time >= self[parent].start_time
        ):
            raise ValueError(
                f"{child} start time must be within {parent} time interval"
            )
        # set the ancestor of the child population
        self[child].ancestors = [parent]
        self.branches.append(Branch(parent, child, time))

    def merge(self, parents, proportions, child, time):
        """
        Add merger event at a given time, where multiple parents contribute to
        a descendant deme, and the parent demes cease to exist at that time.

        :param parents: The ancestral demes.
        :param proportions: Proportions of ancestral demes contributing to descendant.
        :param children: The descendant deme.
        :param time: The time at which merger occurs.
        """
        if self[child].start_time != time:
            raise ValueError(
                f"{child}'s start time must equal admixture time of {time}"
            )
        # for parental populations, we check that their end time is <= merge time
        for parent in parents:
            if self[parent].end_time > time:
                raise ValueError(f"deme {parent} has end time earlier than {time}")
        # if any parent end times are more recent than merge time, we adjust the end
        # and remove epochs that extend beyond that merger time
        for parent in parents:
            if self[parent].end_time < time:
                while self[parent].epochs[-1].end_time < time:
                    if self[parent].epochs[-1].start_time <= time:
                        del self[parent].epochs[-1]
                    else:
                        self[parent].epochs[-1].end_time = time
        # set the ancestors and proportions of the child deme
        self[child].ancestors = parents
        self[child].proportions = proportions
        self.mergers.append(Merge(parents, proportions, child, time))

    def admix(self, parents, proportions, child, time):
        """
        Add admixture event at a given time, where multiple parents contribute to
        a descendant deme, and the parent demes continue to exist beyond that time.

        :param parents: The ancestral demes.
        :param proportions: Proportions of ancestral demes contributing to descendant.
        :param children: The descendant deme.
        :param time: The time at which admixture occurs.
        """
        if self[child].start_time != time:
            raise ValueError(
                f"{child}'s start time must equal admixture time of {time}"
            )
        # for parental populations, we check that their end time is <= admixture time
        for parent in parents:
            if self[parent].end_time > time:
                raise ValueError(f"deme {parent} has end time earlier than {time}")
        # set the ancestors and proportions of the child deme
        self[child].ancestors = parents
        self[child].proportions = proportions
        self.admixtures.append(Admix(parents, proportions, child, time))

    def get_demographic_events(self):
        """
        Loop through successors/predecessors to add splits, branches, mergers,
        and admixtures to the deme graph. If a deme has more than one predecessor,
        then it is a merger or an admixture event, which we differentiate by end and
        start times of those demes. If a deme has a single predecessor, we check
        whether it is a branch (start time != predecessor's end time), or split.

        This is only used when we build a demography from a YAML file, since it
        uses the successors/predecessors that are determined by ancestor relationships.
        """
        splits_to_add = {}
        for c, p in self.predecessors.items():
            if len(p) == 0:
                continue
            elif len(p) == 1:
                if self[c].start_time == self[p[0]].end_time:
                    splits_to_add.setdefault(p[0], set())
                    splits_to_add[p[0]].add(c)
                else:
                    self.branch(p[0], c, self[c].start_time)
            else:
                time_aligned = True
                for deme_from in p:
                    if self[c].start_time != self[deme_from].end_time:
                        time_aligned = False
                if time_aligned is True:
                    self.merge(
                        self[c].ancestors, self[c].proportions, c, self[c].start_time
                    )
                else:
                    self.admix(
                        self[c].ancestors, self[c].proportions, c, self[c].start_time
                    )
        for deme_from, demes_to in splits_to_add.items():
            self.split(deme_from, list(demes_to), self[deme_from].end_time)

    def validate(self):
        """
        Validates the demographic model.
        """
        loads(dumps(self))

    def in_generations(self):
        """
        Return a copy of the demes graph with times in units of generations.
        """
        deme_graph = copy.deepcopy(self)
        deme_graph.time_units = "generations"
        generation_time = self.generation_time
        if generation_time is not None:
            deme_graph.generation_time = None
            for deme in deme_graph.demes:
                for epoch in deme.epochs:
                    epoch.start_time /= generation_time
                    epoch.end_time /= generation_time
            for migration in deme_graph.migrations:
                migration.start_time /= generation_time
                migration.end_time /= generation_time
            for pulse in deme_graph.pulses:
                pulse.time /= generation_time
            deme_graph.splits = []
            deme_graph.branches = []
            deme_graph.mergers = []
            deme_graph.admixtures = []
            deme_graph.get_demographic_events()
        return deme_graph

    def asdict(self):
        """
        Return a dict representation of the deme graph.
        """
        return attr.asdict(self)

    def asdict_compact(self):
        """
        Return a dict representation of the deme graph, with default and
        implicit values removed.
        """
        d = dict(
            description=self.description,
            time_units=self.time_units,
        )
        if self.generation_time is not None:
            d.update(generation_time=self.generation_time)
        if self.doi is not None:
            d.update(doi=self.doi)
        if self.default_Ne is not None:
            d.update(default_Ne=self.default_Ne)

        if self.selfing_rate is not None:
            d.update(selfing_rate=self.selfing_rate)
        if self.cloning_rate is not None:
            d.update(cloning_rate=self.cloning_rate)

        assert len(self.demes) > 0
        d.update(demes=dict())
        for deme in self.demes:
            deme_dict = dict()
            # add ancestors to deme if not None
            if deme.ancestors is not None:
                deme_dict.update(ancestors=deme.ancestors)
                if len(deme.ancestors) > 1:
                    deme_dict.update(proportions=deme.proportions)
                if any([deme.start_time != self[a].end_time for a in deme.ancestors]):
                    deme_dict.update(start_time=deme.start_time)
            else:
                # corner case of no ancestors but finite start time
                if math.isfinite(deme.start_time):
                    deme_dict.update(start_time=deme.start_time)
            # add selfing and cloning rates, if not None
            if deme.selfing_rate is not None:
                if self.selfing_rate is None or (
                    self.selfing_rate is not None
                    and deme.selfing_rate != self.selfing_rate
                ):
                    deme_dict.update(selfing_rate=deme.selfing_rate)
            if deme.cloning_rate is not None:
                if self.cloning_rate is None or (
                    self.cloning_rate is not None
                    and deme.cloning_rate != self.cloning_rate
                ):
                    deme_dict.update(cloning_rate=deme.cloning_rate)

            assert len(deme.epochs) > 0
            e_list = []
            for j, epoch in enumerate(deme.epochs):
                e = dict()
                # end time required for epochs
                e.update(end_time=epoch.end_time)
                e.update(initial_size=epoch.initial_size)
                if epoch.final_size != epoch.initial_size:
                    e.update(final_size=epoch.final_size)
                if epoch.size_function not in ["constant", "exponential"]:
                    e.update(size_function=epoch.size_function)
                if epoch.selfing_rate is not None:
                    if deme.selfing_rate is not None:
                        if epoch.selfing_rate != deme.selfing_rate:
                            e.update(selfing_rate=epoch.selfing_rate)
                    elif self.selfing_rate is not None:
                        if epoch.selfing_rate != self.selfing_rate:
                            e.update(selfing_rate=epoch.selfing_rate)
                    else:
                        e.update(selfing_rate=epoch.selfing_rate)
                if epoch.cloning_rate is not None:
                    if deme.cloning_rate is not None:
                        if epoch.cloning_rate != deme.cloning_rate:
                            e.update(cloning_rate=epoch.cloning_rate)
                    elif self.cloning_rate is not None:
                        if epoch.cloning_rate != self.cloning_rate:
                            e.update(cloning_rate=epoch.cloning_rate)
                    else:
                        e.update(cloning_rate=epoch.cloning_rate)
                e_list.append(e)
            if len(e_list) > 1:
                # if more than one epoch, list all epochs
                deme_dict.update(epochs=e_list)
            else:
                # if a single epoch, don't list as under epochs
                deme_dict.update(initial_size=e_list[0]["initial_size"])
                if "final_size" in e_list[0]:
                    if e_list[0]["final_size"] != e_list[0]["initial_size"]:
                        deme_dict.update(final_size=e_list[0]["final_size"])
                if e_list[0]["end_time"] > 0:
                    deme_dict.update(end_time=e_list[0]["end_time"])
            if deme.description is not None:
                deme_dict.update(description=deme.description)
            d["demes"][deme.id] = deme_dict

        if len(self.migrations) > 0:
            m_dict = collections.defaultdict(list)
            for migration in self.migrations:
                m_dict[(migration.source, migration.dest)].append(
                    dict(rate=migration.rate)
                )
                time_lo, time_hi = self.check_time_intersection(
                    migration.source, migration.dest, None
                )
                if migration.end_time != time_lo:
                    m_dict[(migration.source, migration.dest)][-1].update(
                        end_time=migration.end_time
                    )
                if migration.start_time != time_hi:
                    m_dict[(migration.source, migration.dest)][-1].update(
                        start_time=migration.start_time
                    )
            # collapse into symmetric and asymmetric migrations
            m_symmetric = []
            m_asymmetric = []
            for (source, dest), m_list in m_dict.items():
                # check if there is equal, reverse migration over the same epoch
                if (dest, source) in m_dict:
                    for m in m_list:
                        no_symmetry = True
                        for i, m_compare in enumerate(m_dict[(dest, source)]):
                            if m == m_compare:
                                m_symmetric.append(
                                    dict(demes=[source, dest], rate=m["rate"])
                                )
                                if "start_time" in m:
                                    m_symmetric[-1]["start_time"] = m["start_time"]
                                if "end_time" in m:
                                    m_symmetric[-1]["end_time"] = m["end_time"]
                                # pop the m_compare so we don't repeat it
                                m_dict[(dest, source)].remove(m_compare)
                                no_symmetry = False
                                break
                        if no_symmetry:
                            m_asymmetric.append(
                                dict(source=source, dest=dest, rate=m["rate"])
                            )
                            if "start_time" in m:
                                m_asymmetric[-1]["start_time"] = m["start_time"]
                            if "end_time" in m:
                                m_asymmetric[-1]["end_time"] = m["end_time"]
                else:
                    # all ms in m_list are asymmetric
                    for m in m_list:
                        m_asymmetric.append(
                            dict(source=source, dest=dest, rate=m["rate"])
                        )
                        if "start_time" in m:
                            m_asymmetric[-1]["start_time"] = m["start_time"]
                        if "end_time" in m:
                            m_asymmetric[-1]["end_time"] = m["end_time"]
            migrations_out = {}
            if len(m_symmetric) > 0:
                migrations_out["symmetric"] = m_symmetric
            if len(m_asymmetric) > 0:
                migrations_out["asymmetric"] = m_asymmetric
            if len(migrations_out) > 0:
                d.update(migrations=migrations_out)

        if len(self.pulses) > 0:
            d.update(pulses=[attr.asdict(pulse) for pulse in self.pulses])

        return d



from demes import DemeGraph, Epoch

# The parameters in this file have NOT been verified, and should not be trusted to be
# accurate. For verified demographic models, see stdpopsim implementations.


def zigzag():
    N1 = 1431
    N2 = 14312
    g = DemeGraph(
        description="Sequential exponential increase and decrease of population size.",
        doi="10.1038/ng.3015",
        time_units="years",
        generation_time=30,
    )
    g.deme(
        "ZigZag",
        initial_size=N1,
        epochs=[
            Epoch(end_time=34133.33, initial_size=N1),
            Epoch(end_time=8533.33, final_size=N2),
            Epoch(end_time=2133.33, final_size=N1),
            Epoch(end_time=533.33, final_size=N2),
            Epoch(end_time=133.33, final_size=N1),
            Epoch(end_time=33.33, final_size=N2),
            Epoch(end_time=0, final_size=N2),
        ],
    )
    return g


def gutenkunst_ooa():
    g = DemeGraph(
        description="Gutenkunst et al. (2009) three-population model.",
        doi="10.1371/journal.pgen.1000695",
        time_units="years",
        generation_time=25,
    )
    g.deme("ancestral", end_time=220e3, initial_size=7300)
    g.deme("AMH", ancestors=["ancestral"], end_time=140e3, initial_size=12300)
    g.deme("OOA", ancestors=["AMH"], end_time=21.2e3, initial_size=2100)
    g.deme("YRI", ancestors=["AMH"], initial_size=12300)
    g.deme("CEU", ancestors=["OOA"], initial_size=1000, final_size=29725)
    g.deme("CHB", ancestors=["OOA"], initial_size=510, final_size=54090)
    g.symmetric_migration(demes=["YRI", "OOA"], rate=25e-5)
    g.symmetric_migration(demes=["YRI", "CEU"], rate=3e-5)
    g.symmetric_migration(demes=["YRI", "CHB"], rate=1.9e-5)
    g.symmetric_migration(demes=["CEU", "CHB"], rate=9.6e-5)
    return g


def browning_america():
    g = DemeGraph(
        description="Browning et al. (2011) model of admixture in the Americas.",
        doi="10.1371/journal.pgen.1007385",
        time_units="generations",
        generation_time=25,
    )
    g.deme("ancestral", end_time=5920, initial_size=7310)
    g.deme("AMH", ancestors=["ancestral"], end_time=2040, initial_size=14474)
    g.deme("AFR", ancestors=["AMH"], initial_size=14474)
    g.deme("OOA", ancestors=["AMH"], end_time=920, initial_size=1861)
    g.deme("EUR", ancestors=["OOA"], initial_size=1000, final_size=34039)
    g.deme("EAS", ancestors=["OOA"], initial_size=510, final_size=45852)
    g.deme(
        "ADMIX",
        ancestors=["AFR", "EUR", "EAS"],
        proportions=[1 / 6, 1 / 3, 1 / 2],
        start_time=12,
        initial_size=30000,
        final_size=54664,
    )
    g.symmetric_migration(demes=["AFR", "OOA"], rate=15e-5)
    g.symmetric_migration(demes=["AFR", "EUR"], rate=2.5e-5)
    g.symmetric_migration(demes=["AFR", "EAS"], rate=0.78e-5)
    g.symmetric_migration(demes=["EUR", "EAS"], rate=3.11e-5)
    return g


def jacobs_papuans():
    N_archaic = 13249
    N_DeniAnc = 100
    N_ghost = 8516
    T_Eu_bottleneck = 1659

    g = DemeGraph(
        description="Jacobs et al. (2019) archaic admixture into Papuans",
        doi="10.1016/j.cell.2019.02.035",
        time_units="generations",
        generation_time=29,
    )
    g.deme("ancestral_hominin", end_time=20225, initial_size=32671)
    g.deme(
        "archaic",
        ancestors=["ancestral_hominin"],
        end_time=15090,
        initial_size=N_archaic,
    )

    g.deme("Den1", ancestors=["archaic"], end_time=12500, initial_size=N_DeniAnc)
    g.deme("Den2", ancestors=["Den1"], end_time=9750, initial_size=N_DeniAnc)
    # Altai Denisovan (sampling lineage)
    g.deme("DenAltai", ancestors=["Den2"], initial_size=5083)
    # Introgressing Denisovan lineages 1 and 2
    g.deme("DenI1", ancestors=["Den2"], initial_size=N_archaic)
    g.deme("DenI2", ancestors=["Den1"], initial_size=N_archaic)

    g.deme("Nea", ancestors=["archaic"], end_time=3375, initial_size=N_archaic)
    # Altai Neanderthal (sampling lineage)
    g.deme("NeaAltai", ancestors=["Nea"], initial_size=826)
    # Introgressing Neanderthal lineage
    g.deme("NeaI", ancestors=["Nea"], end_time=883, initial_size=N_archaic)

    g.deme("AMH", ancestors=["ancestral_hominin"], end_time=2218, initial_size=41563)
    g.deme("Africa", ancestors=["AMH"], initial_size=48433)
    g.deme(
        "Ghost1",
        ancestors=["AMH"],
        initial_size=N_ghost,
        epochs=[
            # bottleneck
            Epoch(end_time=2119, initial_size=1394),
            Epoch(end_time=1784, initial_size=N_ghost),
        ],
    )
    g.deme("Ghost2", ancestors=["Ghost1"], end_time=1758, initial_size=N_ghost)
    g.deme("Ghost3", ancestors=["Ghost2"], initial_size=N_ghost)
    g.deme(
        "Papua",
        ancestors=["Ghost1"],
        # bottleneck
        epochs=[
            Epoch(end_time=1685, initial_size=243),
            Epoch(end_time=0, initial_size=8834),
        ],
    )
    g.deme(
        "Eurasia",
        ancestors=["Ghost2"],
        # bottleneck
        epochs=[
            Epoch(end_time=T_Eu_bottleneck, initial_size=2231),
            Epoch(end_time=1293, initial_size=12971),
        ],
    )
    g.deme("WestEurasia", ancestors=["Eurasia"], initial_size=6962)
    g.deme("EastAsia", ancestors=["Eurasia"], initial_size=9025)

    g.symmetric_migration(
        demes=["Africa", "Ghost3"], rate=1.79e-4, start_time=T_Eu_bottleneck
    )
    g.symmetric_migration(demes=["Ghost3", "WestEurasia"], rate=4.42e-4)
    g.symmetric_migration(demes=["WestEurasia", "EastAsia"], rate=3.14e-5)
    g.symmetric_migration(demes=["EastAsia", "Papua"], rate=5.72e-5)
    g.symmetric_migration(
        demes=["Eurasia", "Papua"], rate=5.72e-4, start_time=T_Eu_bottleneck
    )
    g.symmetric_migration(
        demes=["Ghost3", "Eurasia"], rate=4.42e-4, start_time=T_Eu_bottleneck
    )

    g.pulse(source="NeaI", dest="EastAsia", proportion=0.002, time=883)
    g.pulse(source="NeaI", dest="Papua", proportion=0.002, time=1412)
    g.pulse(source="NeaI", dest="Eurasia", proportion=0.011, time=1566)
    g.pulse(source="NeaI", dest="Ghost1", proportion=0.024, time=1853)

    m_Den_Papuan = 0.04
    p = 0.55  # S10.i p. 31
    T_Den1_Papuan_mig = 29.8e3 / g.generation_time
    T_Den2_Papuan_mig = 45.7e3 / g.generation_time
    g.pulse(
        source="DenI1",
        dest="Papua",
        proportion=p * m_Den_Papuan,
        time=T_Den1_Papuan_mig,
    )
    g.pulse(
        source="DenI2",
        dest="Papua",
        proportion=(1 - p) * m_Den_Papuan,
        time=T_Den2_Papuan_mig,
    )

    return g


def IM(num_demes, N, m, time_units="generations", generation_time=1):
    g = DemeGraph(
        description="Isolation with migration.",
        time_units=time_units,
        generation_time=generation_time,
        default_Ne=N,
    )
    for k in range(num_demes):
        g.deme(f"D{k}")
    g.symmetric_migration(demes=[deme.id for deme in g.demes], rate=m)
    return g



import unittest

import demes

# test_examples.py tests importing and exporting YAML files
# these tests are for incremental builds of demographic models, where we
# add demes to an "empty" demography using the demographic event functions


class TestExamples(unittest.TestCase):
    maxDiff = None

    def test_one_deme_multiple_epochs(self):
        # initial size set to 50, add a later deme that has size of 100
        # only need to specify the second deme, the first one is implicit with
        # size equal to 50
        g = demes.DemeGraph(
            description="one deme test", time_units="generations", generation_time=1
        )
        g.deme(
            id="pop",
            initial_size=50,
            epochs=[demes.Epoch(start_time=30, end_time=0, initial_size=100)],
        )
        self.assertEqual(len(g["pop"].epochs), 2)
        self.assertEqual(g["pop"].epochs[0].start_time, float("inf"))
        self.assertEqual(g["pop"].epochs[0].end_time, 30)
        self.assertEqual(g["pop"].epochs[1].start_time, 30)
        self.assertEqual(g["pop"].epochs[1].end_time, 0)
        self.assertEqual(g["pop"].start_time, float("inf"))
        self.assertEqual(g["pop"].end_time, 0)

        # same as above, but start time is not inf
        g = demes.DemeGraph(
            description="one deme test", time_units="generations", generation_time=1
        )
        g.deme(
            id="pop",
            initial_size=50,
            start_time=100,
            epochs=[demes.Epoch(start_time=30, end_time=0, initial_size=100)],
        )
        self.assertEqual(len(g["pop"].epochs), 2)
        self.assertEqual(g["pop"].epochs[0].start_time, 100)

    def test_simple_split(self):
        g = demes.DemeGraph(
            description="split model", time_units="generations", generation_time=1
        )
        g.deme(id="ancestral", end_time=50, initial_size=100)
        g.deme(id="pop1", start_time=50, initial_size=200)
        g.deme(id="pop2", start_time=50, initial_size=300)
        g.split("ancestral", ["pop1", "pop2"], 50)
        self.assertEqual(len(g.splits), 1)
        self.assertTrue(g.splits[0].parent == "ancestral")
        self.assertTrue("pop1" in g.splits[0].children)
        self.assertTrue("pop2" in g.splits[0].children)
        self.assertTrue(g.splits[0].time == 50)
        self.assertTrue("ancestral" in g["pop1"].ancestors)
        self.assertTrue("ancestral" in g["pop2"].ancestors)
        self.assertTrue(g["ancestral"].ancestors is None)

    def test_simple_branch(self):
        g = demes.DemeGraph(
            description="branch model", time_units="generations", generation_time=1
        )
        g.deme(id="ancestral", initial_size=100)
        g.deme(id="pop1", start_time=50, initial_size=200)
        g.branch("ancestral", "pop1", 50)
        self.assertEqual(len(g.branches), 1)
        self.assertTrue(g.branches[0].parent == "ancestral")
        self.assertTrue(g.branches[0].child == "pop1")
        self.assertTrue(g.branches[0].time == 50)
        self.assertTrue("ancestral" in g["pop1"].ancestors)

    def test_simple_merge(self):
        g = demes.DemeGraph(
            description="branch model", time_units="generations", generation_time=1
        )
        g.deme(id="ancestral1", initial_size=100, end_time=10)
        g.deme(id="ancestral2", initial_size=100, end_time=10)
        g.deme(id="child", initial_size=100, start_time=10)
        g.merge(["ancestral1", "ancestral2"], [0.5, 0.5], "child", 10)
        self.assertEqual(len(g.mergers), 1)
        self.assertEqual(g.mergers[0].time, 10)
        self.assertEqual(g.mergers[0].child, "child")
        for anc in ["ancestral1", "ancestral2"]:
            self.assertTrue(anc in g.mergers[0].parents)
        self.assertEqual(g["ancestral1"].end_time, 10)
        self.assertEqual(g["ancestral2"].end_time, 10)
        self.assertEqual(g["child"].start_time, 10)

    def test_merge_that_truncates(self):
        # by calling merge and setting the time, we cut the parental populations
        # at the merge time
        g = demes.DemeGraph(
            description="branch model", time_units="generations", generation_time=1
        )
        g.deme(id="ancestral1", initial_size=100)  # don't set their end times
        g.deme(id="ancestral2", initial_size=100)
        g.deme(id="child", initial_size=100, start_time=10)
        g.merge(["ancestral1", "ancestral2"], [0.5, 0.5], "child", 10)
        self.assertEqual(len(g.mergers), 1)
        self.assertEqual(g.mergers[0].time, 10)
        self.assertEqual(g.mergers[0].child, "child")
        for anc in ["ancestral1", "ancestral2"]:
            self.assertTrue(anc in g.mergers[0].parents)
        self.assertEqual(g["ancestral1"].end_time, 10)
        self.assertEqual(g["ancestral2"].end_time, 10)
        self.assertEqual(g["child"].start_time, 10)

    def test_admixture(self):
        g = demes.DemeGraph(
            description="branch model", time_units="generations", generation_time=1
        )
        g.deme(id="ancestral1", initial_size=100)
        g.deme(id="ancestral2", initial_size=100)
        g.deme(id="child", initial_size=100, start_time=10)
        g.admix(["ancestral1", "ancestral2"], [0.5, 0.5], "child", 10)
        self.assertEqual(g["ancestral1"].end_time, 0)
        self.assertEqual(g["ancestral2"].end_time, 0)
        self.assertEqual(g["child"].end_time, 0)
        self.assertEqual(g["child"].start_time, 10)
        self.assertEqual(len(g.admixtures), 1)
        self.assertEqual(g.admixtures[0].time, 10)
        for anc in ["ancestral1", "ancestral2"]:
            self.assertTrue(anc in g.admixtures[0].parents)
            self.assertTrue(anc in g["child"].ancestors)
